#import sys, getopt
from ribotish.zbio import ribo, plot, io, bam

def help():
  return "Quality control for riboseq data"
def set_parser(parser):
  # basic input options
  parser.add_argument("-b", type=str, dest="ribobampath", required=True, help="Riboseq bam file")
  parser.add_argument("-g", type=str, dest="genepath", help='Gene annotation file')
  parser.add_argument("-o", type=str, dest="output", help="Output data file (default: ribobampath[:-4]+ '_qual.txt')")
  # alt input options
  parser.add_argument('-t', "--tis", action="store_true", help="The data is TIS enriched (for LTM & Harritonine)")
  parser.add_argument("-i", type=str, dest="input", help="Input previous output file, do not read gene file and bam file again")
  parser.add_argument("--geneformat", type=str, default='auto', help="Gene annotation file format (gtf, bed, gpd, gff, default: auto)")
  parser.add_argument("--chrmap", type=str, help="Input chromosome id mapping table file if annotation chr ids are not same as chr ids in bam/fasta files")

  # quality result output
  parser.add_argument("-f", type=str, dest="figpdfpath", help="Output pdf figure file (default: ribobampath[:-4]+ '_qual.pdf')")
  parser.add_argument("-r", type=str, dest="parapath", help="Output offset parameter file (default: ribobampath[:-4]+ '.para.py')")
  # other options
  parser.add_argument("-l", type=int2, dest="lens", default=(25,35), help="Range of tag length (default: 25,35)")
  parser.add_argument("-d", type=int2, dest="dis", default=(-40,20), help="Position range near start codon or stop codon (default: -40,20)")
  parser.add_argument("--bins", type=int, default=20, help="Bins for cds profile (default: 20)")
  parser.add_argument("--nom0", action="store_true", help="Do not consider reads with mismatch at position 0 as a new group")
  parser.add_argument("--th", type=float, default=0.5, help="Threshold for quality (default: 0.5)")
  # Reads filters
  parser.add_argument("--maxNH", type=int, default=1, help="Max NH value allowed for bam alignments (default: 1)")
  parser.add_argument("--minMapQ", type=float, default=1, help="Min MapQ value required for bam alignments (default: 1)")
  parser.add_argument("--secondary", action="store_true", help="Use bam secondary alignments")
  parser.add_argument("--paired", action="store_true", help="Reads are paired end")
  
  parser.add_argument("--colorblind", action="store_true", help="Use a color style readable for color blind people ('#F00078,#00F000,#0078F0')")
  parser.add_argument("--colors", type=strlist, help="User specified Matplotlib accepted color codes for three frames (default: 'r,g,b')")

  parser.add_argument("-p", type=int, dest="numProc", default=1, help="Number of processes (default: 1)")
  parser.add_argument("-v", "--verbose", action="count", help="Increase output verbosity")

def int2(s):
  '''Convert comma seperated string to tuple
  '''
  lst = eval('['+s+']')
  return tuple(map(int, lst))

def strlist(s):
  '''Convert comma seperated file name string to list
  '''
  return s.split(',')

#m0 = True
def run(args):
  '''Main function for quality control
  '''
  global m0
  m0 = True
  if args.nom0 : m0 = False
  ribo.maxNH, ribo.minMapQ, ribo.secondary = args.maxNH, args.minMapQ, args.secondary
  global fbcols
  fbcols = ['r','g','b','r','g','b']
  if args.colorblind:
    fbcols = ['#F00078', '#00F000', '#0078F0'] # ['#FA007D', '#00E800', '#007DFA']
  if args.colors is not None:
    fbcols = args.colors

  if args.input is None :
    minR = 1
    if args.genepath is None : 
      print('Missing -g gene annotation file!')
      exit(1)
    if args.ribobampath is None : 
      print('Missing -b riboseq bam file!')
      exit(1)
    if args.output is None : 
      args.output = args.ribobampath[:-4] + '_qual.txt'
    if args.chrmap is not None :
      chrmap = {}
      for lst in io.splitIter(args.chrmap):
        chrmap[lst[0]] = lst[1]
        chrmap[lst[1]] = lst[0]
      bam.chrmap = chrmap
    #fa.chrmap = chrmap

    # read data
    results = ribo.lendis(args.genepath, args.ribobampath, lens = args.lens, dis = args.dis, minR = minR, m0 = m0, paired = args.paired,
                          cdsBins = args.bins, numProc = args.numProc, verbose = args.verbose, geneformat = args.geneformat)
    # write quality data
    outfile = open(args.output, 'w')
    s = 0
    for ad in results :
      s += ad.size()
      ad.write(outfile)
    print('Counted reads: {}'.format(s))
    outfile.close()
    args.input = args.output
  if args.figpdfpath is None : 
    args.figpdfpath = args.ribobampath[:-4] + '_qual.pdf'
  if args.parapath is None : 
    args.parapath = args.ribobampath + '.para.py'
  # get quality plot and quality parameter file
  qualityPlot(args)

#fbcols = ['r','g','b','r','g','b']
#if args.colorblind:
  #fbcols = ['#FF008C', '#00FF00', '008CFF']

def qualityPlot(args):
  ''' Quality plot and offset parameters
  '''
  global m0
  if args.figpdfpath is None : args.figpdfpath = args.ribobampath[:-4] + '_qual.pdf'
    #print('Need to provide -f output pdf figure file name!')
    #exit(1)
  elif args.verbose : print('Quality pdf figure will be saved to {}'.format(args.figpdfpath))
  if args.parapath is None : args.parapath = args.ribobampath + '.para.py'
    #print('Need to provide -r output offset parameter file name!')
    #exit(1)
  elif args.verbose : print('Offset parameter will be saved to {}'.format(args.parapath))
  offdict = {}
  infile = open(args.input,'r')

  lines = [l.rstrip() for l in infile]
  lendis = eval(lines[0])
  dis1 = eval(lines[1])
  dis2 = eval(lines[2])
  disf = eval(lines[3])
  disc = eval(lines[4])
  if m0 and len(lines) >= 10 : 
    lendism0 = eval(lines[5])
    dis1m0 = eval(lines[6])
    dis2m0 = eval(lines[7])
    disfm0 = eval(lines[8])
    discm0 = eval(lines[9])
    m0 = True
    offdict['m0'] = {}
  else : m0 = False

  fig = plot.figure(figsize = (16, 16))
  k = list(lendis.keys())
  k.sort()
  #x = list(range(*args.dis))
  if m0 : gs0 = plot.matplotlib.gridspec.GridSpec(1, 2, top = 0.92, bottom = 0.86)
  else : gs0 = plot.matplotlib.gridspec.GridSpec(1, 1, top = 0.92, bottom = 0.86)

  ax = plot.subplot(gs0[0, 0])
  x1 = [l - 0.4 for l in lendis.keys()]
  ax.bar(x1, lendis.values(), alpha=0.6)
  formatax(ax)
  ty = nearest(max(lendis.values()+[1]), up=False)
  ax.set_yticks([0, ty])
  ax.set_yticklabels([0,numk(ty)])
  if m0 : ax.set_title("5' end match RPFs\n\nRPF length distribution")
  else : ax.set_title("RPFs\n\nRPF length distribution")
  if m0 : 
    ax = plot.subplot(gs0[0, 1])
    x1 = [l - 0.4 for l in lendism0.keys()]
    ax.bar(x1, lendism0.values(), alpha=0.6)
    formatax(ax)
    ty = nearest(max(lendism0.values()+[1]), up=False)
    ax.set_yticks([0, ty])
    ax.set_yticklabels([0,numk(ty)])
    ax.set_title("5' end mismatch RPFs\n\nRPF length distribution")

  if m0 : gs = plot.matplotlib.gridspec.GridSpec(len(k), 12, top = 0.80, bottom = 0.05, left = 0.07, right = 0.98, wspace=0.7, hspace=0.3)
  else : gs = plot.matplotlib.gridspec.GridSpec(len(k), 6, top = 0.80, bottom = 0.05, left = 0.1, right = 0.95, wspace=0.35, hspace=0.3)

  for i, l in enumerate(k):
    plot4(gs, i, l, disf, dis1, dis2, disc, offdict, start = 0, args=args)
    if m0 : plot4(gs, i, l, disfm0, dis1m0, dis2m0, discm0, offdict['m0'], start = 6, args=args)
  plot.save(args.figpdfpath) 

  parafile = open(args.parapath, 'w') 
  ribo.write_off_para(parafile, offdict)
  total = sum(lendis.values())
  if m0: total += sum(lendism0.values())
  n = nt = 0
  for l in offdict :
    if l in disf : 
      n += sum(disf[l])
      if args.tis : nt += dis1[l][-args.dis[0] - offdict[l]]
  if m0 :
    for l in offdict['m0'] :
      n += sum(disfm0[l])
      if args.tis : nt += dis1m0[l][-args.dis[0] - offdict['m0'][l]]
  if args.verbose : 
    print('Total RPF counts in mRNA: {}'.format(total))
    print('Effective RPF counts: {}'.format(n))
    if args.tis : print('Effective TIS counts: {}'.format(nt))

def nearest(n, up = True, step = 10, levels = range(1, 10)):
  lasti = i = e = 1
  while n >= i :
    for l in levels:
      i = e * l
      if n <= i : break
      lasti = i
    if n <= i : break
    e *= step
  if up : return i
  else : return lasti
def numk(n):
  step = 1000
  s = ['k', 'M', 'G', 'T']
  i = -1
  while n >= step :
    n /= step
    i += 1
    if i >= len(s) - 1 : break
  if i >= 0 : return str(n) + s[i]
  else : return str(n)
def frange(start=0,stop=1,step=1):
  lst = []
  i = start
  while i < stop:
    lst.append(i)
    i += step
  return lst
def stdisplot(ax, x, y, lab = '', hali = 'left', vali = 'top', frame = 3, f0 = 0, color = ['r','g','b'], size='medium', tcol = 'k'):
  for i in range(frame):
    i0 = (f0 + i) % frame
    ax.bar(x[i0::frame], y[i0::frame], width = 0.6, color = color[i], edgecolor = color[i],alpha=0.6)
  if hali == 'left' : tx = min(x)
  else : tx = max(x)
  ty = nearest(max(y+[1]), up=False) ##
  ax.text(tx ,max(y+[1])*0.7 , lab,horizontalalignment=hali,verticalalignment=vali, size=size, color = tcol)
  ax.spines['right'].set_visible(False)
  ax.spines['top'].set_visible(False)
  ax.spines['left'].set_visible(False)
  ax.tick_params(direction='in', top = 'off', right = 'off')
  ax.set_yticks([0, ty])
  ax.set_yticklabels([0,numk(ty)])
def cdsplot(ax, disc):
  m = 1
  for i in range(3):
    ax.plot(disc[i], color = fbcols[i])
    mi = max(disc[i])
    if m < mi : m = mi
  formatax(ax)
  ty = nearest(m, up=False)
  ax.set_yticks([0, ty])
  ax.set_yticklabels([0,numk(ty)])
  ax.set_xticks([])
def formatax(ax):
  ax.spines['right'].set_visible(False)
  ax.spines['top'].set_visible(False)
  ax.spines['left'].set_visible(False)
  ax.tick_params(direction='in', top = 'off', right = 'off')


def plot4(gs, i, l, disf, dis1, dis2, disc, offdict, args, start = 0):
  y0 = disf[l]
  ys = sum(y0)
  if ys > 0 : y = [float(n)/ys for n in y0]
  else : y = y0
  ax0 = plot.subplot(gs[i, start])
  ax0.bar(frange(0.2, len(y)), y, width = 0.6, color = fbcols, edgecolor = fbcols, alpha=0.6)
  use, frame, txt = ribo.quality(y, threshold = args.th)
  ax0.text(3.1 ,0.9 ,txt, horizontalalignment= 'right',verticalalignment='top', color='k')
  formatax(ax0)
  #ax0.set_xticks(range(4))
  ax0.set_xticks([])
  ax0.set_yticks(frange(stop=1.1, step=0.5))
  if start == 0 : 
    ax0.set_ylabel(str(l)+'nt',rotation=0,size='x-large')
    if m0 : ax0.yaxis.set_label_coords(-0.9, 0.4)
    else : ax0.yaxis.set_label_coords(-0.4, 0.4)

  if args.tis : 
    use, tisframe, tistxt, mp = ribo.TISquality(dis1[l], dis = args.dis, threshold = args.th)
    if tisframe != frame : use = False
    #print use, frame, tistxt
  txt = ''
  if i == 0 : txt = "offset:"
  if use : 
    offset, th = ribo.get_offset(dis1[l], dis = args.dis, frame = frame, tis = args.tis)
    if offset is not None :
      txt += str(offset)
      offdict[l] = offset
    #else : txt = ''
  else : txt += "NA"
  tcol = 'k'
  if use : tcol = fbcols[frame]
  ax1 = plot.subplot(gs[i, start+1:start+3])
  x = list(range(*args.dis))
  stdisplot(ax1, x, dis1[l], lab = txt, hali = 'left', f0 = -args.dis[0], color=fbcols, tcol = tcol)
  
  if use and not args.tis : 
    plot.hlines(th, -18, -6, color=tcol, linestyles ='dashed')
  if args.tis : 
    #mp = ribo.getTIS(dis1[l], [d1,d2])
    ax1.text(mp+1 ,max(dis1[l]+[1])*0.7 , tistxt, horizontalalignment='left',verticalalignment='top', size='medium', color = tcol)
  ax2 = plot.subplot(gs[i, start+3:start+5])
  
  stdisplot(ax2, x, dis2[l], lab = '', hali = 'right', f0 = -args.dis[0], color=fbcols, tcol = tcol)
  ax3 = plot.subplot(gs[i, start+5])
  cdsplot(ax3, disc[l])
  if args.tis : 
    bins = len(disc[l][tisframe])
    mean = sum(disc[l][tisframe]) / bins
    if mean == 0 : enrich = 0
    else : enrich = dis1[l][mp - args.dis[0]] / mean
    if i == 0 : enrtxt = 'TIS\nenrich\n%.1f' % enrich
    else : enrtxt = '%.1f' % enrich
    ax3.text(bins ,ax3.get_ylim()[1] , enrtxt, horizontalalignment='right',verticalalignment='top', size='medium', color = tcol)
  if i == 0 : ax0.set_title('Frame distr.\n')
  if i == 0 : ax1.set_title('RPF near start codon\n')
  if i == 0 : ax2.set_title('RPF near stop codon\n')
  if i == 0 : ax3.set_title('CDS Profile\n')
  if i == gs._nrows - 1 : ax0.set_xlabel('Frame in codon', size='medium')
  if i == gs._nrows - 1 : ax1.set_xlabel('Dist. from start codon (nt)', size='medium')
  if i == gs._nrows - 1 : ax2.set_xlabel('Dist. from stop codon (nt)', size='medium')
  #if i == 0 and m0 : ax1.text(0, 0, "5' end match RPFs", ha='left', va='bottom')

if __name__=="__main__":
  import sys, argparse
  p = argparse.ArgumentParser()
  set_parser(p)
  if len(sys.argv)==1:
    print(p.print_help())
    exit(0)
  run(p.parse_args())



'''
Fasta format file processing
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''

from os.path import isfile

def faIter(file, protein = False):
  id = sq = ""
  for l in file:
    l = l.strip()
    if l == '' : continue
    if l[0] == '>':
      if id != '':
        yield (id,sq)
      sq = ''
      id = l[1:]
    else:
      if not protein : l = l.replace('U','T').replace('u','t')
      sq += l
  yield (id, sq)

def rc(seq):
  '''reverse complement
  '''
  comps = {'A':"T", 'C':"G", 'G':"C", 'T':"A",
       'B':"V", 'D':"H", 'H':"D", 'K':"M",
       'M':"K", 'R':"Y", 'V':"B", 'Y':"R",
       'W':'W', 'N':'N', 'S':'S'}
  return ''.join([comps[x] for x in seq.upper()[::-1]])

def changechr(chr):
  '''change between two chr versions
  '''
  if chr.isdigit() or chr in ('X','Y','M'): return 'chr' + chr
  elif chr == 'MT' : return 'chrM'
  elif chr == 'chrM' : return 'MT'
  elif chr[0:3] == 'chr' : return chr[3:]
  else : return chr

chrmap = {}

class Faidx:
  '''index for genome fasta file
  '''
  def __init__(self, fid, length = 0, pos = 0, ls = 0, ll = 0):
    '''
    ls : sequence length in one line
    ll : line length
    '''
    self.id = fid
    self.length = length
    self.pos = pos
    self.ls = ls
    self.ll = ll
    self.is_endls = False
    self.is_endll = False
  def __str__(self):
    return "%s\t%d\t%d\t%d\t%d" % (self.id, self.length, self.pos, self.ls, self.ll)
  
  def updatels(self, ls):
    if ls <= 0 : raise Exception("Empty sequence!")
    if self.ls == 0 : self.ls = ls
    elif self.is_endls : raise Exception("Sequence has ended: %d" % (ls))
    elif ls > self.ls :
      raise Exception("Length error: %d -> %d" % (self.ls, ls))
    elif ls < self.ls : self.is_endls = True
  def updatell(self, ll):
    #print ll
    if ll <= 0 : raise Exception("Empty line!")
    if self.ll == 0 : self.ll = ll
    elif self.is_endll : raise Exception("Sequence has ended: %d" % (ll))
    elif ll > self.ll :
      raise Exception("Length error: %d -> %d" % (self.ll, ll))
    elif ll < self.ll : self.is_endll = True
      #print 'End: ', self.ll, ll

class Fa:
  '''indexed genome fasta file
  '''
  def __init__(self, fapath, verbose = False):
    self.file = open(fapath, 'r')
    idxpath = fapath + '.fai'
    if isfile(idxpath) : self.load_idx(idxpath)
    else : 
      if verbose : print('Making fasta index for {}...'.format(fapath))
      self.make_idx(idxpath)
  def make_idx(self, idxpath):
    self.file.seek(0)
    self.idxarr = []
    pos, flen = 0, 0
    for l in self.file:
      pos += len(l)
      if l[0] == '>' : 
        if flen > 0 : self.idxarr[-1].length = flen
        fid = l.strip().split()[0][1:]
        #print fid
        self.idxarr.append(Faidx(fid, pos = pos))
        flen = 0
      else : 
        ll, ls = len(l), len(l.strip())
        flen += ls
        self.idxarr[-1].updatels(ls)
        self.idxarr[-1].updatell(ll)
        #if ll > self.idxarr[-1][4] : self.idxarr[-1][4] = ll
    self.idxarr[-1].length = flen
    outfile = open(idxpath, 'w')
    self.idx = {}
    for i, idx in enumerate(self.idxarr):
      outfile.write('{}\n'.format(idx)) # print >>outfile, idx
      self.idx[idx.id] = idx
    outfile.close()
  def load_idx(self, idxpath):
    self.idx = {}
    for l in open(idxpath):
      lst = l.strip().split('\t')
      if len(lst) < 5 : continue
      idx = Faidx(lst[0])
      idx.length, idx.pos, idx.ls, idx.ll = list(map(int, lst[1:5]))
      self.idx[idx.id] = idx
  def get_chrname(self, chr) : #, chrmap = chrmap):
    #sprint(chrmap)
    if chr in self.idx : return chr
    mt = {'M':'MT', 'MT':'M'}
    if chr in mt and mt[chr] in self.idx: return mt[chr]
    if chr in chrmap :
      chr1 = chrmap[chr]
      if chr1 in self.idx : return chr1
      else : 
        chr2 = changechr(chr1)
        if chr2 in self.idx : return chr2
    chr1 = changechr(chr)
    if chr1 == chr : return None
    if chr1 in self.idx : return chr1
    elif chr1 in chrmap : 
      chr2 = chrmap[chr1]
      if chr2 in self.idx : return chr2
    return None
  def seekpos(self, fid, p):
    idx = self.idx[fid]
    pos = idx.pos # chr start
    pos += int(p / idx.ls) * idx.ll # line start
    pos += p % idx.ls # base start
    return pos

  def get_pos(self, fid, start = 0, stop = -1, length = -1):
  #def fetch(self, fid, start = 0, stop = -1, length = -1):
    chr = self.get_chrname(fid)
    if chr is None : 
      print("fa id {} not found in fasta file!".format(fid))
      return None, None, None, None
    fid = chr
    if start < 0 : start = start % self.idx[fid].length
    if length < 0 : 
      if stop < 0 : stop = stop % self.idx[fid].length
      length = stop - start
    return fid, start, stop, length

  def fetch(self, fid, start = 0, stop = -1, length = -1):
    fid, start, stop, length = self.get_pos(fid, start, stop, length)
    if fid is None: return None
    if length <= 0 : return ''
    self.file.seek(self.seekpos(fid, start))
    seq = ''
    idx = self.idx[fid]
    for l in self.file:
      l = l.strip()
      if length > len(l) :
        seq += l
        length -= len(l)
      else : 
        seq += l[0:length]
        break
    return seq
  def __iter__(self):
    for fid in self.idx:
      yield fid
    #return self.idx.itervalues()
  def exonSeq(self, e):
    s = self.fetch(e.chr, start = e.start, stop = e.stop)
    if e.strand == '-': s = rc(s)
    return s
  def transSeq(self, trans):
    s = []
    for e in trans.exons: s.append(self.exonSeq(e))
    return ''.join(s)

  def intervalSeq(self, fid, regions, strand='+'):
    chr = self.get_chrname(fid)
    if chr is None :
      print("fa id {} not found in fasta file!".format(fid))
      return None
    s = []
    for itv in regions:
      s.append(self.fetch(chr, itv[0], itv[1]))
    sq = ''.join(s)
    if strand == '-': sq = rc(sq)
    return sq





'''
Intervals processing
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''

def is_overlap(i1, i2):
  return i1[0] < i2[1] and i1[1] > i2[0]

class Interval:
  '''intervals in 1 dimentional axis
  all intervals are supposed to be [start, end) and start <= end
  '''
  def __init__(self, start = None, stop = None, id = '', itvs = [], adj_merge=True):
    self.id = id
    self.lst = [] # list of (start, stop) tuples
    self.adj_merge = adj_merge
    if itvs == []:
      if start is not None and start <= stop : self.lst.append((start, stop))
    else :
      self.lst += itvs
      #for itv in itvs:
        #self.lst.append(itv[:])
    self.check()
  def check(self):
    for itv in self.lst[:]:
      if itv[0] > itv[1] : self.lst.remove(itv)
    self.lst.sort()
    dl = [] # delete list
    j = 0
    for i in range(1, len(self.lst)):
      if self.lst[i][0] < self.lst[j][1] or (self.adj_merge and self.lst[i][0] == self.lst[j][1]) :
        if self.lst[j][1] < self.lst[i][1] : self.lst[j] = (self.lst[j][0], self.lst[i][1]) # self.lst[j][1] = self.lst[i][1]
        dl.append(i)
      else : j = i
    for i in dl[::-1]: del self.lst[i]
  def __len__(self):
    return len(self.lst)
  def rlen(self):
    l = 0
    for itv in self.lst: l += itv[1] - itv[0]
    return l
  def __repr__(self):
    s = 'interval '+ self.id + ':'
    if len(self.lst) == 0 : s += ' empty!'
    for itv in self.lst:
      s += ' [{}, {})'.format(itv[0],itv[1])
    return s
  def __getitem__(self, i): 
    return self.lst[i]
  def __add__(self, other): # do not change self
    new = Interval(itvs = self, adj_merge=self.adj_merge)
    new.lst += list(other)
    #for itv in other:
      #new.lst.append(itv[:])
    new.check()
    return new
  def add(self, other):
    return self + other
  def add_itv(self, itv, check=True): # input single interval, change the object
    self.lst.append(tuple(itv))
    if check: self.check()
    return self
  def sub_itv(self, itv):
    if itv[0] > itv[1] : return self
    lst = []
    for i in self.lst:
      if is_overlap(i, itv) : 
        if i[0] < itv[0] : lst.append((i[0],itv[0]))
        if i[1] > itv[1] : lst.append((itv[1],i[1]))
      else : lst.append(i)
    self.lst = lst
    self.check()
    return self
  def __sub__(self, other): # to be optimized, not tested
    '''substract other intervals, self and other should both be sorted
    '''
    lst = []
    j = 0
    for itv in self.lst:
      while j < len(other) and other[j][1] <= itv[0] : j += 1
      j1 = j
      while j1 < len(other) and other[j1][0] < itv[1] : j1 += 1
      if j1 == j : lst.append(itv) # no overlap interval
      else : # overlap with j...j1-1
        if itv[0] < other[j][0] : lst.append((itv[0], other[j][0]))
        for i in range(j, j1-1) : lst.append((other[i][1], other[i+1][0]))
        if itv[1] > other[j1-1][1] : lst.append((other[j1-1][1], itv[1]))
    new = Interval(itvs = lst, adj_merge=self.adj_merge)
    return new
    #new = Interval(itvs = self)
    #for itv in other:
      #new.sub_itv(itv[:])
    #new.check()
    #return new
  def sub(self, other):
    return self - other
  def ints_itv(self, itv): # intersect 
    if itv[0] > itv[1] : 
      self.lst = []
      return self
    lst = []
    for i in self.lst:
      if is_overlap(i, itv): lst.append((max(i[0], itv[0]), min(i[1], itv[1])))
    self.lst = lst
    self.check()
    return self
  def intersect(self, other): # optimized
    '''intersect intervals, self and other should both be sorted
    '''
    if len(other) == 0 or len(self) == 0 : return Interval()
    os = other[0][0]
    osi, c = self.nearest(os)
    lst = []
    j = 0
    #for itv in self.lst:
    for i in range(osi, len(self.lst)) :
      itv = self.lst[i]
      while j < len(other) and other[j][1] <= itv[0] : j += 1
      j1 = j
      while j1 < len(other) and other[j1][0] < itv[1] :
        lst.append((max(itv[0], other[j1][0]), min(itv[1], other[j1][1])))
        j1 += 1
      if j >= len(other) : break
    new = Interval(itvs = lst, adj_merge=self.adj_merge)
    return new
    #return (self + other) - (self - other) - (other - self) # hehe
  def nearest(self, p): 
    ''' nearest interval index in self.lst, return index, cmp(p, lst[index])
    '''
    l = len(self.lst)
    if l == 0 : return -1, 1
    lasti = i = l // 2
    i0, i1 = 0, l
    while i0 < i1 : 
      if self.lst[i][0] == p : break
      if self.lst[i][0] < p : i0 = i
      else : i1 = i
      lasti = i
      i = (i0 + i1) // 2
      if i == lasti : break
    if self.lst[i][0] <= p <= self.lst[i][1] : return i, 0 # inside
    if self.lst[i][0] > p : return i, -1 # p is smaller than ith interval, only when p < self.start
    if self.lst[i][1] < p : return i, 1 # p is greater than ith interval
    
  def is_inside(self, p, left = True, right = False, strand = '+') : # is p inside interval
    if len(self.lst) == 0 : return False
    i, c = self.nearest(p)
    if c != 0 : return False
    if self.lst[i][0] < p < self.lst[i][1] : return True
    if strand == '-' : left, right = right, left
    if left and self.lst[i][0] == p : return True
    if right and self.lst[i][1] == p : return True
    return False
  def nearestPos(self, p, strand = '+', downstream = True):
    i, c = self.nearest(p)
    if c == 0 : return p
    larger = True
    if strand != '-' and downstream == False : larger = False
    if strand == '-' and downstream == True :  larger = False
    if larger : 
      if c > 0 : plus = 1
      else : plus = 0
      if i + plus < len(self.lst) : return self.lst[i+plus][0]
      else : return None
    else : 
      if c > 0 : minus = 0
      else : minus = 1
      if i - minus >= 0 : return self.lst[i-minus][1]
      else : return None
  def nearestStop(self, p, strand = '+', downstream = True):
    i, c = self.nearest(p)
    if c != 0 : return p
    larger = True
    if strand != '-' and downstream == False : larger = False
    if strand == '-' and downstream == True :  larger = False
    if larger : return self.lst[i][1]
    else : return self.lst[i][0]
  def num_iter(self, start = None, step = 1, strand = '+'):
    if self.rlen() <= 0 : return
    if start is None : start = self.end5(strand)
    i = start
    if strand == '-' : step = -step
    s1, s2 = self.start, self.stop
    while s1 <= i <= s2 : 
      if self.is_inside(i, strand=strand) : yield i
      i += step
    #for i in range(start, self.stop, step):
      #if self.is_inside(i) : yield i
  @property
  def start(self):
    if self.is_empty() : return None
    else : return self.lst[0][0]
  @property
  def stop(self):
    if self.is_empty() : return None
    else : return self.lst[-1][1]
  def end5(self, strand = '+'):
    if strand != '-' : return self.start
    else : return self.stop
  def end3(self, strand = '+'):
    if strand != '-' : return self.stop
    else : return self.start
  def is_empty(self):
    return len(self.lst) == 0
  def is_compatible(self, other, mis = 0) :
    p1 = max(self.start, other.start)
    p2 = min(self.stop, other.stop)
    if p2 - p1 <= mis : return True # not overlap
    si = self.intersect([(p1, p2)])
    oi = other.intersect([(p1, p2)])
    
    m = si.sub(oi).rlen()
    if m > mis : return False
    m += oi.sub(si).rlen()
    if m > mis : return False
    return True
  def contradict(self, other) :
    p1 = max(self.start, other.start)
    p2 = min(self.stop, other.stop)
    if p2 <= p1 : return None # not overlap, empty
    si = self.intersect([(p1, p2)])
    oi = other.intersect([(p1, p2)])
    return si.sub(oi), oi.sub(si)
  def genome_pos(self, p, strand = '+', bias = 1): # exon only
    if len(self) == 0: return None
    m = self.rlen() # cdna_length()
    if p < 0 or p > m: return None
    if strand == '-': 
      p = m - p
      bias = 1 - bias
    if p == 0 : return self.start
    if p == m: return self.stop
    p1 = p
    for itv in self.lst:
      l = itv[1] - itv[0]
      if l - p1 >= bias:
        return p1 + itv[0] # e.start
      else:
        p1 -= l
    return None
  def cdna_pos(self, p, strand = '+', strict = False):
    '''if strict is True, the 3' end of exon will be considered as not in the transcript,
    if strict is False, 3' end of exon will be considered as start of the next exon, 
    or transcript end (self.cdna_length()) if in the last exon.
    '''
    if len(self.lst) == 0 : return None
    i, c = self.nearest(p)
    if c != 0 : return None
    if strict:
      if strand != '-' and p == self.lst[i][1]: return None
      if strand == '-' and p == self.lst[i][0]: return None
    pos = sum([self.lst[j][1] - self.lst[j][0] for j in range(i)])
    pos += p - self.lst[i][0]
    if strand == '-': 
      pos = self.rlen() - pos
    return pos
 
def trans2interval(t, start = 0, stop = None):
  '''generate intervals for a transcript
  start, stop are cDNA positions
  '''
  i = Interval(id = t.chr, itvs = [(e.start, e.stop) for e in t.exons])
  if start == 0 and stop is None : return i
  s1 = t.end5
  if start > 0 : s1 = t.genome_pos(start)
  s2 = t.end3
  if stop is not None : s2 = t.genome_pos(stop)
  if s2 is None : s2 = t.end3 # = t.genome_pos(start), t.genome_pos(stop)
  itv = [s1, s2]
  itv.sort()
  i.ints_itv(itv)
  return i
  #itv = Interval(id = t.chr)
  #for e in t.exons:
    #itv.lst.append((e.start, e.stop))
  #itv.check()
  #return itv

def cds_region_trans(t, cds1 = None, cds2 = None):
  '''generate CDS regions in 3 frames of chromosome
  '''
  cr = [Interval() for i in range(3)]
  if cds1 is None : cds1 = t.cds_start(cdna = True) 
  if cds2 is None : cds2 = t.cds_stop(cdna = True)
  if cds1 is None : return cr
  if cds2 is None : 
    tl = t.cdna_length()
    cds2 = cds1 + (tl - cds1) // 3 * 3
  #print cds1, cds2
  if cds2 - cds1 == 0 : return cr
  wrong = False
  if (cds2 - cds1) % 3 > 0 : 
    print('Wrong CDS annotation: %s %s %d %d %d' % (t.gid, t.id, cds1, cds2, t.cdna_length()))
    wrong = True
    return cr ## Wrong CDS annotation
  thick = [t.genome_pos(cds1, 1), t.genome_pos(cds2, 0)]
  thick.sort()
  ts1, ts2 = thick #t.thick_start, t.thick_stop
  orf = t(start = ts1, stop = ts2)
  exons = t.exons[:]
  exons.sort()
  if wrong : 
    for e in t.cds : 
      df = int(e.frame) # Use frame annotation
      if t.strand == '+' : df = - df
      f = (e.end5 - df) % 3
      cr[f].lst.append((e.start, e.stop))
  else : 
    df = 0
    for e in exons:
      for ei in e.intersect(orf):
        f = (ei.start - df) % 3
        cr[f].lst.append((ei.start, ei.stop))
        df = (len(ei) + df) % 3
      #print f, df, ei
    if df != 0 : 
      print('Remain df is not 0 ! %s %s %d %d' % (t.gid, t.id, cds1, cds2))
      return [Interval() for i in range(3)]
  for r in cr : r.check()
  return cr
def cds_region_gene(g):
  cr = [Interval() for i in range(3)]
  for t in g.trans:
    tcr = cds_region_trans(t)
    for i in range(3):
      cr[i] += tcr[i]
  return cr

def allTransRegions(trans):
  regions = {}
  for t in trans :
    transitv = trans2interval(t)
    if t.chr not in regions : regions[t.chr] = Interval(id = t.chr)
    regions[t.chr].lst += transitv.lst
  for chr in regions: regions[t.chr].check()
  return regions

def str2interval(s, blocksep = ',', rangesep = '-', id = ''):
  itvs = [map(int, b.split(rangesep)) for b in s.split(blocksep)]
  return Interval(itvs=itvs, id=id)





'''
Bam file and reads processing
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''
import pysam
from . import interval

def changechr(chr):
  '''change between two chr versions
  '''
  if chr.isdigit() or chr in ('X','Y','M'): return 'chr' + chr
  elif chr == 'MT' : return 'chrM'
  elif chr == 'chrM' : return 'MT'
  elif chr[0:3] == 'chr' : return chr[3:]
  else : return chr

chrmap = {}
    
class Bamfile(pysam.Samfile):
  def __repr__(self):
    return 'pysam.Samfile '+self.filename
  def get_chrname(self, chr) : #, chrmap = chrmap):
    #sprint(chrmap)
    if chr in self.references : return chr
    if chr in chrmap :
      chr1 = chrmap[chr]
      if chr1 in self.references : return chr1
      else : 
        chr2 = changechr(chr1)
        if chr2 in self.references : return chr2
    chr1 = changechr(chr)
    if chr1 == chr : return None
    if chr1 in self.references : return chr1
    elif chr1 in chrmap : 
      chr2 = chrmap[chr1]
      if chr2 in self.references : return chr2
    return None
  def fetch_reads(self, chr, start, stop, maxNH = None, minMapQ = None, secondary = False, paired = False) : #, multiple_iterators=False):
    chr0, chr = chr, self.get_chrname(chr)
    #if chr not in self.references : 
      #chr = changechr(chr)
    if chr is None : 
      print("chr {} not found in Bamfile!".format(chr0))
      raise StopIteration
    rds = self.fetch(reference=chr, start=start, end=stop) #, multiple_iterators=multiple_iterators)
    r1, r2 = {}, {}
    for read in rds:
      try: 
        if maxNH is not None and read.get_tag('NH') > maxNH : continue
      except: pass
      if not secondary and read.is_secondary : continue
      if minMapQ is not None and read.mapping_quality < minMapQ : continue
      if paired and read.is_paired :
        id = read.query_name
        if id[-2:] in ('\1', '\2', '#1', '#2') : id = id[:-2] ##
        if read.is_read1 : 
          if id in r2 : 
            yield Bam(read, self, read2 = r2[id])
            del r2[id]
          else : r1[id] = read
        elif read.is_read2 : 
          if id in r1 : 
            yield Bam(r1[id], self, read2 = read)
            del r1[id]
          else : r2[id] = read
      else : yield Bam(read, self)
      #yield r
    if paired :
      for id in r1 : 
        if r1[id].mate_is_unmapped : continue
        if r1[id].next_reference_id != r1[id].reference_id : continue
        if r1[id].is_reverse != (not r1[id].mate_is_reverse) : continue
        if abs(r1[id].next_reference_start - r1[id].reference_start) > 1e6 : continue
        #print(id, 'r1')
        read = self.mate(r1[id])
        yield Bam(r1[id], self, read2 = read)
    #if len(r2) > 0 :
      for id in r2 : 
        if r2[id].mate_is_unmapped : continue
        if r2[id].next_reference_id != r2[id].reference_id : continue
        if r2[id].is_reverse != (not r2[id].mate_is_reverse) : continue
        if abs(r2[id].next_reference_start - r2[id].reference_start) > 1e6 : continue
        #print(id, 'r2')
        read = self.mate(r2[id])
        yield Bam(read, self, read2 = r2[id])

class Bam():#AlignedRead
  
  def __init__(self, read, bamfile, read2 = None):
    self.read = read
    self.ref = bamfile.references
    self.lens = bamfile.lengths
    self.read2 = read2
  #def __getattr__(self,attr):
    #a=getattr(self,attr)
    #return a
  
  @property
  def chr(self): 
    #print self.ref['chr1']
    return self.ref[self.read.tid]
  @property
  def start(self):
    return self.read.reference_start
  @property
  def fragment_start(self):
    if self.read2 is None : return self.read.reference_start
    else :
      if self.is_reverse() : return self.read2.reference_start
      else : return self.read.reference_start
    #except : return self.read.pos
  @property
  def stop(self):
    return self.read.reference_end
    #return int(self.read.reference_end)
  @property
  def fragment_stop(self):
    if self.read2 is None : return self.read.reference_end
    else :
      if self.is_reverse() : return self.read.reference_end
      else : return self.read2.reference_end
  @property
  def id(self): 
    return self.read.query_name
  @property
  def score(self): 
    try:
      return self.read.get_tag("AS")
    except:
      return 0.0
  @property
  def strand(self): 
    if self.read.is_reverse: return '-'
    else: return '+'
  @property
  def cigar(self): 
    try : return self.read.cigartuples
    except : return self.read.cigar
  @property
  def weight(self): 
    return 1
  def __len__(self): 
    return self.read.reference_length
  @property
  def length(self):
    return len(self)
  def fragment_length(self): 
    if self.read2 is None : return self.read.reference_length
    else : return abs(self.read.template_length)
  def __repr__(self):
    if self.read2 is None : return "Bam AlignedSegment {} {}".format(self.id, str(self))
    else : return "Bam paired AlignedSegment {} {}".format(self.id, str(self))
  def __str__(self):
    return "%s:%d-%d:%s" % (self.chr, self.fragment_start, self.fragment_stop, self.strand)

  def get_tag(self, k):
    return self.read.get_tag(k)
  def cdna_length(self): 
    return self.read.query_alignment_length
    #else : return self.fragment_length()

  def fragment_length(self):
    if self.read2 is None : return self.cdna_length()
    r1 = interval.Interval(itvs = self.read.get_blocks())
    r2 = interval.Interval(itvs = self.read2.get_blocks())
    ri = r2.intersect(r1)
    ril = ri.rlen()
    if ril > 0 : 
      if not self.is_reverse() :
        r1s = r1.intersect(interval.Interval(r1.start, ri.stop))
        r2s = r2.intersect(interval.Interval(ri.stop, r2.stop))
      else : 
        r1s = r1.intersect(interval.Interval(ri.start, r1.stop))
        r2s = r2.intersect(interval.Interval(r2.start, ri.start))
      return r1s.rlen() + r2s.rlen()
    else : # there may be problems here
      l = max(r1.start, r2.start) - min(r1.stop, r2.stop) # uncovered length, may be not accurate
      if l < 0 : l = 0 ###
      l += self.read.query_alignment_length + self.read2.query_alignment_length 
      
      return l
  def center(self): #Middle point, NEED REVISE!!
    return (self.start+self.stop)/2.0
  
  def is_reverse(self): 
    return self.read.is_reverse
  def is_paired(self):
    return self.read2 is not None
  def is_pair_gapped(self):
    if self.read2 is None : return False
    if not self.is_reverse : return self.read.reference_end < self.read2.reference_start
    else : return self.read2.reference_end < self.read.reference_start
  @property
  def end5(self): #5' end
    if self.read.is_reverse: return self.stop
    else : return self.start
  @property
  def end3(self): #3' end
    if self.read.is_reverse: return self.start
    else : return self.stop
  @property
  def fragment_end5(self): 
    if self.read2 is None : return self.end5
    if self.is_reverse: return self.fragment_stop
    else : return self.fragment_start
  @property
  def fragment_end3(self): 
    if self.read2 is None : return self.end3
    if self.is_reverse: return self.fragment_start
    else : return self.fragment_stop
  def __getattr__(self, name):
    if name == '__class__': return str
    else: return self.read.__getattribute__(name)
    #try : return getattr(self, name)
    #except AttributeError : return getattr(self.read, name)
  def cdna_pos(self, p, strict = False): 
    '''NOT READY because of insertions and deletions
    '''
    if p < self.start or p > self.stop:
      return None
    #p1 = p - self.start
    pos = 0
    for e in self.exons:
      if e.start <= p <= e.stop:
        if self.is_reverse():
          pos += e.stop - p
        else:
          pos += p - e.start
        return pos
      else:
        pos += len(e)
    return None
  
  def genome_pos(self, p, bias = 1):
    '''if bias is 1, the splice junction will be mapped to 5' end of downstream exon,
    if bias is 0, the splice junction will be mapped to 3' end of the upstream exon.
    Do not support paired end read2
    '''
    if p is None : return None
    m = self.cdna_length()
    if p < 0 or p > m: return None
    if p == 0 : return self.end5
    if p == m: return self.end3
    p1 = p
    if self.is_reverse():
      p1 = self.cdna_length() - p
      if bias == 0: bias = 1
      else: bias = 0
    pos = self.start
    for ctype, l in self.cigar:
      if ctype in [0,7,8]: # M: alignment match, =: sequence match, X: sequence mismatch
        if l - p1 >= bias:
          pos += p1
          return pos
        else:
          pos += l
          p1 -= l
      elif ctype in [1,]:#4]: # I: insertion to the reference, S: soft clipping
        if l - p1 >= bias: return pos
        else: p1 -= l
      elif ctype in [2,3]: # D: deletion from the reference, N: skipped region from the reference
        pos += l
    return None
  @property
  def introns(self):
    from . import bed
    s = []
    p = 0
    pos = self.start
    i = 1
    for ctype, l in self.cigar:
      if ctype in [0,7,8]:
        pos += l
        p += l
      elif ctype in [1]:#,4]:
        p += l
      elif ctype in [2]:
        pos += l
      elif ctype in [3]:
        b = bed.bed6([self.chr,pos,pos + l,self.id+"_intron"+str(i),p,self.strand])
        s.append(b)
        pos += l
        i += 1
    if self.is_reverse():
      return s[::-1]
    return s
  @property
  def exons(self): # NOT TESTED!! Insertions and deletions are skipped
    from . import bed
    s = []
    p = 0
    pos, d = self.start, 0
    i = 0
    for ctype, l in self.cigar:
      if ctype in [0,7,8]:
        d += l
        p += l
      elif ctype in [1]:#,4]:
        p += l
      elif ctype in [2]:
        d += l
      elif ctype in [3]:
        b = bed.bed6([self.chr,pos,pos + d,self.id+"_exon"+str(i),p,self.strand])
        s.append(b)
        pos += d + l
        i += 1
    b = bed.bed6([self.chr,pos,pos + d,self.id+"_exon"+str(i),p,self.strand])
    s.append(b)
    if self.is_reverse():
      return s[::-1]
    return s
  def isCompatible(self, transitv, mis = 0):
    '''if compatible with given transcript, allow some (mis) incompatible bases
    '''
    if transitv.__class__.__name__ != 'Interval' : 
      transitv = interval.trans2interval(transitv)
    if self.read2 is None : return self.readCompatible(transitv, mis = mis)
    else : return self.readCompatible(transitv, mis = mis) and self.readCompatible(transitv, read = self.read2, mis = mis)
  def readCompatible(self, transitv, read = None, mis = 0):
    if read is None : read = self.read
    ri = interval.Interval(itvs = read.get_blocks())
    return ri.is_compatible(transitv, mis = mis)
    #ti = ri.intersect(transitv)
    #if ri.rlen() - ti.rlen() <= mis : return True
    #return False
    
  def isCompatible0(self, transitv, mis = 0):
    
    #from . import interval
    if transitv.__class__.__name__ != 'Interval' : 
      transitv = interval.trans2interval(transitv)
    r1 = self.start
    r2 = r1 + sum([l for ctype, l in self.cigar])
    ri = interval.Interval(r1, r2) # full reads region
    ti = ri.intersect(transitv) #interval.trans2interval(trans)
    start, stop = ti.start, ti.stop
    m = 0
    pos = self.start
    for ctype, l in self.cigar:
      if ctype in [0,7,8]: # M: alignment match, =: sequence match, X: sequence mismatch
        p1, p2 = max(pos, start), min(pos + l, stop)
        pos += l
        if p1 < p2 : 
          ci = interval.Interval(p1, p2)
          itv = ci.intersect(transitv)
          m += p2 - p1 - itv.rlen()
      elif ctype in [1]:#,4]: # I: insertion to the reference, S: soft clipping
        if start < pos < stop : m += l
      elif ctype in [2,3]: # D: deletion from the reference, N: skipped region from the reference
        p1, p2 = max(pos, start), min(pos + l, stop)
        pos += l
        if p1 < p2 : 
          ci = interval.Interval(p1, p2)
          itv = ci.intersect(transitv)
          m += itv.rlen()
      #print(ctype, l, m)
      if m > mis : return False
    return True
  def is_compatible(self, trans, mis = 0): # old version, slower
    m = 0
    last = -1
    for i in range(self.cdna_length()):
      pos = self.genome_pos(i, bias = 1)
      if pos < trans.start or pos > trans.stop : continue
      i2 = trans.cdna_pos(pos, strict = True)
      #if i2 is None : m += 1
      #else :
      if i2 is not None :
        if last >= 0 :
          if i2 != last + 1 :
            m += min(i, self.cdna_length() - i, abs(i2 - last - 1))
        else : m += min(i, i2)
        last = i2
      #print(i, i2, last, m)
      if m > mis : return False
    #print self, self.cigar, m
    if m > mis : return False
    else : return True
  def is_compatible0(self, trans = None, introns = [], mis = 0): # incorrect version...
    #if trans != None : exons = trans.exons
    if trans is not None : introns = trans.introns
    m = 0
    ris = self.introns
    if len(ris) == 0 : 
      for intr in introns:
        o = self.read.get_overlap(intr.start, intr.stop)
        if o > 0 : m += o
        if m > mis : return False
    else :
      for i in range(len(introns)):
        if not ris[0].is_upstream(introns[i]) : break
      c = True
      for ri in ris:
        if ri != introns[i] : 
          c = False
          o = self.read.get_overlap(introns[i].start, introns[i].stop)
          p = int(ri.score)
          m += max(o, min(p, self.cdna_length() - p))
          if m > mis : return False
        i += 1
    return True
  def is_inside(self, trans, mis = 0):
    '''only inside, not necessarily compatible
    '''
    l = self.cdna_length()
    al = 0
    for e in trans.exons:
      al += self.read.get_overlap(e.start, e.stop)
    if l - al > mis : return False ## 
    else: return True
  def is_m0(self): 
    '''if mismatch at 5' end
    '''
    if self.strand == '+' : 
      if self.read.get_tag('MD')[0] == '0' : return True # mismatch at 0
      #elif self.cigar[0][0] == 4 : return True
      #else : return False
    elif self.read2 is None :
      if self.read.get_tag('MD')[-1] == '0' : 
        if not self.get_tag('MD')[-2].isdigit() : return True
      #elif self.cigar[-1][0] == 4 : return True
      #else : return False
    else :
      if self.read2.get_tag('MD')[-1] == '0' : 
        if not self.read2.get_tag('MD')[-2].isdigit() : return True
    return False
def compatible_bam_iter(bamfile, trans, mis = 0, sense = True, maxNH = None, minMapQ = None, secondary = False): 
  '''compatible version of transReadsIter, slightly different
  '''
  chr = bamfile.get_chrname(trans.chr)
  if chr is None : raise StopIteration
    #print("chr {} not found in Bamfile!".format(chr0))
    #raise StopIteration
  #chr = trans.chr
  #if chr not in bamfile.references : 
      #chr = changechr(chr)
      #if chr not in bamfile.references : raise StopIteration
  rds = bamfile.fetch(reference=chr, start=trans.start, end=trans.stop)#, multiple_iterators=False)
  #introns = trans.introns
  for r in rds:
    read = Bam(r, bamfile)
    if sense and read.strand != trans.strand:
      #print read.id + " not sense"
      continue
    try: 
      if maxNH is not None and read.read.get_tag('NH') > maxNH : continue
    except: pass
    if not secondary and r.is_secondary : continue
    if minMapQ is not None and read.read.mapping_quality < minMapQ : continue
    o = read.read.get_overlap(trans.start, trans.stop)
    if o < read.cdna_length() - mis: 
      continue
    if read.is_compatible(trans = trans, mis = mis) :
      yield read
    #else:
      #print read.id, b, read.cdna_length()
def transReadsIter(bamfile, trans, compatible = True, mis = 0, sense = True, maxNH = None, minMapQ = None, secondary = False, paired = False):
  '''fetch reads from transcript exons
  '''
  chr = bamfile.get_chrname(trans.chr)
  if chr is None : 
    #print("chr {} not found in Bamfile!".format(trans.chr))
    raise StopIteration
  #chr = trans.chr
  #if chr not in bamfile.references : 
      #chr = changechr(chr)
      #if chr not in bamfile.references : raise StopIteration
  if compatible : 
    from . import interval
    transitv = interval.trans2interval(trans)
  used = {}
  #trans.exons = trans.exons ##
  for e in trans.exons : 
    rds = bamfile.fetch_reads(chr=chr, start=e.start, stop=e.stop, maxNH = maxNH, minMapQ = minMapQ, secondary = secondary, paired = paired)#, multiple_iterators=False)
    for read in rds: #yield read
      #read = Bam(r, bamfile)
      if (read.id, read.fragment_start) in used : continue
      if sense and read.strand != trans.strand : continue
      #try: 
        #if maxNH is not None and read.read.get_tag('NH') > maxNH : continue
      #except: pass
      #if not secondary and read.is_secondary : continue
      #if minMapQ is not None and read.read.mapping_quality < minMapQ : continue
      #o = read.read.get_overlap(trans.start, trans.stop)
      #if o < read.cdna_length() - mis: continue
      if read.fragment_start <= e.start or read.fragment_stop >= e.stop : 
        used[(read.id, read.fragment_start)] = 1 # for reads across exons
      if compatible and not read.isCompatible(transitv = transitv, mis = mis) : continue
      yield read

def end5(r):
  return r.end5

class BamLoad:
  def __init__(self, bampath, transRegions = None, posFunc = end5, numProc = 1, pool = None, maxNH = None, minMapQ = None, secondary = False, verbose = False):
    self.bamloads = {}
    self.bampath = bampath
    if bampath is None : return
    self.regions = transRegions
    #if verbose : print('Loading reads...')
    #import itertools
    if numProc >= 1 and pool is None :
      from multiprocessing import Pool
      pool = Pool(processes = numProc)
    if self.regions is not None :
      paras = [(bampath, chr, self.regions[chr], '.', posFunc, maxNH, minMapQ, secondary) for chr in self.regions]
      #for chr in self.regions:
        #self.bamloads[chr] = BamLoadChr(bampath, chr, self.regions[chr], '.', numProc, pool, posFunc, maxNH, minMapQ, secondary)
    else :
      bamfile = Bamfile(bampath)
      paras = [(bampath, chr, None, '.', posFunc, maxNH, minMapQ, secondary) for chr in bamfile.references]
    #if numProc <= 1 : load_iter = itertools.imap(getBamLoadChr, paras)
    #else : 
      #load_iter = pool.imap_unordered(getBamLoadChr, paras, chunksize = 1)
    loads = pool.map(getBamLoadChr, paras, chunksize = 1)
    #results = []
    #if self.regions is not None :
      #for chr in self.regions :
        #results.append(pool.apply_async(BamLoadChr, (bampath, chr, self.regions[chr], '.', posFunc, maxNH, minMapQ, secondary)))
    #else : 
      #bamfile = Bamfile(bampath)
      #for chr in bamfile.references :
        #results.append(pool.apply_async(BamLoadChr, (bampath, chr, None, '.', posFunc, maxNH, minMapQ, secondary)))
      #n = len(bamfile.references)
      #paras = [bampath]*n, bamfile.references, [None]*n, ['.']*n, [posFunc]*n, [maxNH]*n, [minMapQ]*n, [secondary]*n
      #paras = [(bampath, chr, None, '.', posFunc, maxNH, minMapQ, secondary) for chr in bamfile.references]
    #loads = map(BamLoadChr, paras)
    
    
    #loads = [pool.apply_async(BamLoadChr, (bampath, chr, self.regions[chr], '.', posFunc, maxNH, minMapQ, secondary)) for chr in self.regions]
    #loads = pool.imap_unordered(get_BamLoadChr, paras, chunksize = 1)
    #pool.close()
    #pool.join()
    for blc in loads:
      self.bamloads[blc.chr] = blc
    #pool.close()
    #pool.join()
    #for blc in loads : self.bamloads[blc.chr] = blc

  def merge(self, other):
    for chr in other.bamloads:
      if chr not in self.bamloads: self.bamloads[chr] = BamLoadChr(None, chr)#other.bamloads[chr]
      self.bamloads[chr].merge(other.bamloads[chr])
  def transCounts(self, trans, compatible = False, mis = 2):
    return self.bamloads[trans.chr].transCounts(trans, compatible = compatible, mis = mis)

def getBamLoadChr(args):
  bampath, chr, region, strand, posFunc, maxNH, minMapQ, secondary = args
  return BamLoadChr(bampath, chr, region, strand, posFunc, maxNH, minMapQ, secondary)

class BamLoadChr:
  def __init__(self, bampath, chr, region = None, strand = '.', posFunc = end5, maxNH = None, minMapQ = None, secondary = False, paired = False):
    self.chr = chr
    self.strand = strand
    self.bampath = bampath
    self.region = region
    if strand == '.' : self.data = {'+':{}, '-':{}}
    else : self.data = {strand:{}}
    if bampath is None : return
    bamfile = Bamfile(bampath)
    chr = bamfile.get_chrname(chr)
    if chr is None : #raise StopIteration
    #if chr not in bamfile.references : 
      #chr = changechr(chr)
      #if chr not in bamfile.references : 
      print("chr {} not found in Bamfile!".format(self.chr))
      return
    self.chr = chr
    if region is None : region = [[None, None]]
    for start, stop in region : 
      for r in bamfile.fetch_reads(chr, start, stop, maxNH = maxNH, minMapQ = minMapQ, secondary = secondary, paired = paired):
        if r.strand not in self.data : continue
        p = posFunc(r)
        #print p
        if start is not None and p < start : continue
        if stop is not None and p > stop : continue
        if p not in self.data[r.strand] : self.data[r.strand][p] = {}
        if r.read2 is None : key =tuple(r.read.get_blocks()), #r.start, tuple(r.cigar)
        else : key = tuple(r.read.get_blocks()), tuple(r.read2.get_blocks())
        #print key
        if key not in self.data[r.strand][p] : self.data[r.strand][p][key] = 0
        self.data[r.strand][p][key] += 1
  def __repr__(self):
    return 'BamLoad {}:{}:{}'.format(self.bampath, self.chr, self.strand)
  def transCounts(self, trans, compatible = False, mis = 2):
    if self.strand != '.' and self.strand != trans.strand :
      print('Strand not match: {}, {}'.format(self, trans.id))
      return []
    d = self.data[trans.strand]
    if compatible : 
      from . import interval
      transitv = interval.trans2interval(trans)
    if trans.strand != '-' : step = 1
    else : step = -1
    cnts = [0] * trans.cdna_length()
    for e in trans.exons:
      p = e.end5
      i = trans.cdna_pos(p)
      for j in range(len(e)):
        if p in d :
          if compatible :
            for key in d[p]:
              if self._compatible(key, transitv, mis = mis) : cnts[i] += d[p][key]
          else : cnts[i] = sum(d[p].values())
        i += 1
        p += step
    return cnts
  def _compatible(self, key, transitv, mis = 2):
    '''if compatible with given transcript, allow some (mis) incompatible bases
    '''
    for blocks in key : 
      ri = interval.Interval(itvs = blocks)
      if not ri.is_compatible(transitv, mis = mis) : return False
    return True
  def _compatible0(self, key, transitv, mis = 2): # key structure changed
    
    from . import interval
    r1 = key[0]
    r2 = r1 + sum([l for ctype, l in key[1]])
    ri = interval.Interval(r1, r2) # full reads region
    ti = ri.intersect(transitv) #interval.trans2interval(trans)
    start, stop = ti.start, ti.stop
    m = 0
    pos = key[0] # self.start
    for ctype, l in key[1] : #self.cigar:
      if ctype in [0,7,8]: # M: alignment match, =: sequence match, X: sequence mismatch
        p1, p2 = max(pos, start), min(pos + l, stop)
        pos += l
        if p1 < p2 : 
          ci = interval.Interval(p1, p2)
          itv = ci.intersect(ti)
          m += p2 - p1 - itv.rlen()
      elif ctype in [1]:#,4]: # I: insertion to the reference, S: soft clipping
        if start < pos < stop : m += l
      elif ctype in [2,3]: # D: deletion from the reference, N: skipped region from the reference
        p1, p2 = max(pos, start), min(pos + l, stop)
        pos += l
        if p1 < p2 : 
          ci = interval.Interval(p1, p2)
          itv = ci.intersect(ti)
          m += itv.rlen()
      if m > mis : return False
    return True
  def merge(self, other):
    for s in other.data:
      if s not in self.data : self.data[s] = {}
      for p in other.data[s]:
        if p not in self.data[s] : self.data[s][p] = {}
        for key in other.data[s][p]:
          if key not in self.data[s][p] : self.data[s][p][key] = 0
          self.data[s][p][key] += other.data[s][p][key]
        



'''
Bed format annotation processing
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''
class Bed3:
  ''' Bed3 class for bed3 annotation format, father of Bed6
  '''
  Header=('chr','start','stop')
  Format=(str,int,int)
  n=3
  def __init__(self, x, bin = False): #Fit all bed
    '''init for all bed format
    x: can be a bed format string, Bed object, dict or list
    '''
    if type(x)==str:
      l = x.strip().split('\t')
      if bin: l[0:1] = []
    elif type(x) == type(self) : l = [getattr(x, i) for i in self.Header]
    elif type(x)==dict: l=[x[i] for i in self.Header]
    else: l = x
    lst=[]
    for i in range(self.n):
      try: lst.append(self.Format[i](l[i]))
      except: 
        if i in (6, 7) : lst.append(lst[1]) # blockstart/stop
        elif i in (8,) : lst.append('0') # rgb
        elif i == 9 : lst.append(1)
        elif i == 10 : lst.append((lst[2]-lst[1],))
        elif i == 11 : lst.append((0,))
        elif i in (3,4,5) : lst.append('.')
        else : raise ValueError
    self.items = tuple(lst)
    self.symbol = self.gid = self.tid = self.id ###
    self.genetype = ''
    if type(x) == dict:
      for i in x :
        if i not in self.Header : setattr(self, i, x[i])
  def __str__(self): #Bed3 and Bed6
    return '\t'.join(map(str,self.items))
  def __repr__(self): #All bed
    return 'Bed'+str(self.n)+' object:\n'+str(self)+'\n'
  def short_str(self):
    return "%s:%d-%d:%s" % (self.chr, self.start, self.stop, self.strand)
  @property
  def chr(self): #All bed
    return self.items[0]
  @property
  def start(self): #All bed
    return self.items[1]
  @property
  def stop(self): #All bed
    return self.items[2]
  @property
  def id(self): #Bed3 only
    return "noname"
  @property
  def tid(self):
    return self.id
  @property
  def gid(self):
    return self.tid
  @property
  def symbol(self):
    return self.gid
  @property
  def genetype(self):
    return ""
  @property
  def score(self): #Bed3 only
    return 0.0
  @property
  def strand(self): #Bed3 only
    return "."
  def __len__(self): #All bed
    return self.stop - self.start
  @property
  def length(self):
    return len(self)
  @property
  def end(self): #All bed
    return self.stop
  @property
  def dict(self): #All bed
    return dict([(self.Header[i],self.items[i]) for i in range(self.n)])
  @property
  def end5(self): #5' end, all bed
    if self.strand != '-' : return self.start
    else : return self.stop
  @property
  def end3(self): #3' end, all bed
    if self.strand != '-' : return self.stop
    else : return self.start

  def is_reverse(self): #All bed
    return self.strand=='-'
  def bed(self): #Bed copy, Bed3 only
    '''make a copy of self
    '''
    return Bed3(self)
  def __getitem__(self,i): #All bed
    return self.items[i]
  def __cmp__(self, other): #All bed
    return cmp(self.chr,other.chr) or cmp(self.start,other.start) or cmp(self.stop,other.stop)
  def __sub__(self, other):
    return sub(self, other)
  def intersect(self, other):
    return intersect(self, other)
  def union(self, other):
    return union(self, other)
  def headerline(self,sep='\t'):# Header string, fit all bed
    sep=str(sep)
    return sep.join(self.Header)
  def head(self,n=10): #Bed3 only
    if n>len(self):
      n=len(self)
    return Bed3([self.chr,self.start,self.start+n])
  def tail(self,n=10): #Bed3 only
    if n>len(self):
      n=len(self)
    return Bed3([self.chr,self.stop-n,self.stop])
  def __call__(self,**args): 
    '''New bed that modified from self. For all bed
    '''
    cp=self.bed()
    if args=={}: return cp
    d=self.dict
    for k in args: d[k] = args[k]
      #else: assert 0, k+' not defined!\n'
    cp.__init__(d)
    return cp
  def center(self): #Middle point, Bed3 and Bed6
    return (self.start+self.stop)/2.0
  def cdna_length(self): #Bed3 and Bed6
    return len(self)
  @property
  def exons(self): #Bed3 and Bed6
    l=[]
    e = self(id=self.id+"_Exon_1")
    e.symbol, e.gid, e.tid = self.symbol, self.gid, self.tid
    l.append(e)
    return l
  @property
  def introns(self): # No intron, empty
    return []
  @property
  def trans(self):
    return [self]
  @property
  def thick_start(self):
    return self.start
  @property
  def thick_stop(self):
    return self.start
  @property
  def itemRgb(self):
    return 0
  @property
  def blockCount(self):
    return 1
  @property
  def blockSizes(self):
    return (len(self),)
  @property
  def blockStarts(self):
    return (0,)

  def merge_trans(self):
    return self
  def is_contain(self, p, strict = False): # all bed
    if self.start <= p <= self.stop :
      if not strict : return True
      else : return p != self.end3
    else : return False
  def is_upstream(self, p, strict = False) : # all bed
    if self.strand != '-' : return p < self.start
    else : return p > self.stop
  def is_downstream(self, p, strict = False) : # all bed
    if self.strand != '-' : 
      if strict : return p >= self.stop
      else : return p > self.stop
    else : 
      if strict : return p <= self.start
      else : return p < self.start
  def is_exon(self, p, strict = False): #Bed3 and Bed6
    return self.is_contain(p)
  def is_intron(self, p, strict = False): #False, Bed3 and Bed6
    return False
  def is_sense(self,other): #In same strand? All bed
    return self.strand==other.strand
  def cdna_pos(self, p, strict = False): #Bed3 and Bed6
    if self.is_contain(p, strict): return abs(p-self.end5)
    else: return None
  def genome_pos(self, p, bias = 1): #Bed3 and Bed6
    m = self.cdna_length()
    if p < 0 or p > m: return None
    if p == 0 : return self.end5
    if p == m: return self.end3
    if not self.is_reverse() : return self.start + p
    else : return self.stop - p
  def is_overlap(self, other):
    if(self.chr != other.chr) : return False
    if (self.stop <= other.start) : return False
    if (other.stop <= self.start) : return False
    return True

  def is_compatible(self, other, ignoreStrand = False): # test needed
    if not ignoreStrand :
      if self.strand != other.strand : return False
    if not self.is_overlap(other) : return True
    start = max(self.start, other.start)
    stop = min(self.stop, other.stop)
    int1 = []
    for i in self.introns:
      if i.start <= start : return False
      if i.stop >= stop : return False
      int1.append(i)
    int2 = []
    for i in other.introns:
      if i.stop <= start : continue
      if i.start <= start : return False
      if i.stop >= stop : return False
      int2.append(i)
    if len(int1) != len(int2) : return False
    int1.sort()
    int2.sort()
    for i in range(len(int1)):
      if int1[i] != int2[i] : return False
    return True

class Bed6(Bed3):
  Header=('chr','start','stop','id','score','strand')
  Format=(str,int,int,str,str,str)
  n=6
  #def __init__(self,x):
  #def __str__(self):
  @property
  def id(self): #Bed6 and Bed12
    return self.items[3]
  @property
  def score(self): #Bed6 and Bed12
    return self.items[4]
  @property
  def strand(self): #Bed6 and Bed12
    return self.items[5]
  @property
  def anti_strand(self):
    if self.strand == '+': return '-'
    elif self.strand == '-': return '+'
    else : return '.'
  def bed(self): #Only Bed6, make a copy
    return Bed6(self)
  
  def head(self, n=10): #Bed6 and Bed12
    if n>len(self): n=len(self)
    if self.strand != '-':
      start=self.start
      stop=start+n
    else:
      stop=self.stop
      start=stop-n
      if start<0: start=0
    return Bed6([self.chr,start,stop,self.id+'_Head_'+str(n),self.score,self.strand])
  
  def tail(self,n=10): #Bed6 and Bed12
    if n>len(self): n=len(self)
    if self.strand == '-':
      start=self.start
      stop=start+n
    else:
      stop=self.stop
      start=stop-n
      if start<0: start=0
    return Bed6([self.chr,start,stop,self.id+'_Tail_'+str(n),self.score,self.strand])
  
  #def is_upstream(self, other): #other is upstream of self
    #if self.strand == '-' : return other.start >= self.stop
    #else : return other.stop <= self.start

def com2tup(s): #comma string to tuple
  if type(s)==tuple: return s
  if type(s)==list: return tuple(s)
  return tuple(map(int,s.strip().strip(',').split(',')))
def tup2com(t): #tuple to comma string
  if type(t)==str: return t
  return ','.join(map(str,t))+','

class Bed12(Bed6):
  
  Header=('chr','start','stop','id','score','strand',"thick_start","thick_stop","itemRgb","blockCount","blockSizes","blockStarts")
  Format=(str,int,int,str,str,str,int,int,com2tup,int,com2tup,com2tup)
  n=12
  @property
  def thick_start(self): #Bed12
    return self.items[6]
  @property
  def thick_stop(self):
    return self.items[7]
  #@property
  def cds_start(self, cdna = False): #Bed12
    if self.items[6] == self.items[7] : return None ##
    if self.strand != '-' : s = self.items[6]
    else : s = self.items[7]
    if not cdna : return s
    else : return self.cdna_pos(s)
  #@property
  def cds_stop(self, cdna = False):
    if self.items[6] == self.items[7] : return None ##
    if self.strand != '-' : s = self.items[7]
    else : s = self.items[6]
    if not cdna : return s
    else : return self.cdna_pos(s)
  @property
  def itemRgb(self):
    return self.items[8]
  @property
  def blockCount(self):
    return self.items[9]
  @property
  def blockSizes(self):
    return self.items[10]
  @property
  def blockStarts(self):
    return self.items[11]

  @property
  def blockStops(self):
    return tuple(map(lambda x,y:x+y,self.blockStarts,self.blockSizes))
  def blockStop(self,i): 
    return self.blockStarts[i]+self.blockSizes[i]
  
  def __str__(self): #Bed string, Bed12 only
    l=list(self.items)
    l[8]=tup2com(l[8])
    l[10]=tup2com(l[10])
    l[11]=tup2com(l[11])
    return '\t'.join(map(str,l))

  def genePredStr(self, geneName = True, extended = False):
    lst = []
    if geneName : lst += [self.gid]
    lst += [self.id, self.chr, self.strand, self.start, self.stop, self.thick_start, self.thick_stop, self.blockCount]
    exonStarts = [x + self.start for x in self.blockStarts]
    exonStops = [x + self.start for x in self.blockStops]
    lst += [tup2com(exonStarts), tup2com(exonStops)]
    if extened : 
      lst += [self.name2, tup2com(self.exonFrames), self.cdsStartStat, self.cdsEndStat]
    return '\t'.join(map(str, lst))
  def bed(self): #Bed copy, Bed12 only
    return Bed12(self)
  
  def cdna_length(self): #Bed12
    #l = 0
    #for i in self.blockSizes: l += i
    return sum(self.blockSizes)
  
  def center(self): #Middle point, Bed12 only
    len=self.cdna_length()/2.0
    for i in range(self.blockCount):
      len-=self.blockSizes[i]
      if len < 0 : break
    return self.start+self.blockStarts[i]+self.blockSizes[i]+len
  @property
  def exons(self): #Bed12
    if hasattr(self, '_exons') : return self._exons
    a=[]
    if self.strand=="-": step, j = -1, self.blockCount
    else: step, j = 1, 1
    for i in range(self.blockCount):
      start = self.start + self.blockStarts[i]
      end = start + self.blockSizes[i]
      id=self.id+"_Exon_"+str(j)
      j+=step
      e = Bed6((self.chr,start,end,id,self.score,self.strand))
      e.symbol, e.gid, e.tid = self.symbol, self.gid, self.tid ##
      a.append(e)
    if self.strand=="-": self._exons = a[::-1]
    else: self._exons = a
    return self._exons
  @property
  def introns(self):
    if hasattr(self, '_introns') : return self._introns
    a=[]
    if self.strand=="-": step, j = -1, self.blockCount - 1
    else : step, j = 1, 1
    for i in range(self.blockCount-1):
      start=self.start+self.blockStarts[i]+self.blockSizes[i]
      end=self.start+self.blockStarts[i+1]
      id=self.id+"_Intron_"+str(j)
      j+=step
      e = Bed6((self.chr,start,end,id,self.score,self.strand))
      e.symbol, e.gid, e.tid = self.symbol, self.gid, self.tid ##
      a.append(e)
    if self.strand=="-": self._introns = a[::-1]
    else : self._introns = a
    return self._introns

  def abs2relative(self):
    '''change absolute blockStart values to relative values
    '''
    blockstarts = [st - self.start for st in self.blockStarts]
    return self(blockStarts = blockstarts)

  def cdna_pos(self, p, strict = False):
    '''if strict is True, the 3' end of exon will be considered as not in the transcript,
    if strict is False, 3' end of exon will be considered as start of the next exon, 
    or transcript end (self.cdna_length()) if in the last exon.
    '''
    if not self.is_contain(p, strict) : return None
    p1 = p - self.start # genome relative posistion
    pos = 0 # cDNA pos
    l=list(range(self.blockCount))
    if self.is_reverse(): l = l[::-1]
    for i in l:
      if self.blockStarts[i]<=p1<=self.blockStop(i):
        if strict :
          if self.strand != '-' and p1 == self.blockStop(i): return None
          if self.strand == '-' and p1 == self.blockStarts[i]: return None
        if self.is_reverse(): pos+=self.blockStop(i)-p1
        else: pos+=p1-self.blockStarts[i]
        return pos
      else: pos+=self.blockSizes[i]
      #print(pos)
    else: return None

  def genome_pos(self, p, bias=1):
    '''if bias is 1, the splice junction will be mapped to 5' end of downstream exon,
    if bias is 0, the splice junction will be mapped to 3' end of the upstream exon.
    '''
    m = self.cdna_length()
    if p < 0 or p > m: return None
    if p == 0 : return self.end5
    if p == m: return self.end3
    p1=p
    pos=self.start
    l=list(range(self.blockCount))
    if self.is_reverse(): l=l[::-1]
    for i in l:
      if self.blockSizes[i]-p1>=bias:
        if self.is_reverse(): pos+=self.blockStop(i)-p1
        else: pos+=p1+self.blockStarts[i]
        return pos
      else: p1-=self.blockSizes[i]
    else: return None 
  
  #def cdna_type(p):
    
  def type(self, p, genome = True):
    ''' 5UTR, CDS or 3UTR
    '''
    if genome == False:
      pg = self.genome_pos(p)
      return self.type(pg)
    if type(p) != int : return None
    if p > self.stop or p < self.start : return None
    if p < self.thick_start:
      if self.is_reverse(): return "3UTR"
      else: return "5UTR"
    elif  p > self.thick_stop:
      if self.is_reverse(): return "5UTR"
      else: return "3UTR"
    else: return "CDS"

  def is_exon(self, p, strict = False): #If in exon, Bed12
    if not self.is_contain(p, strict = strict): return False
    l=list(range(self.blockCount))
    if self.is_reverse() : l=l[::-1]
    p1=p-self.start
    for i in l:
      if self.blockStarts[i]<=p1<=self.blockStop(i):
        if strict :
          if self.strand != '-' and p1 == self.blockStop(i): return False
          if self.strand == '-' and p1 == self.blockStarts[i]: return False
        return True
    else: return False
    #return self.is_contain(p)
  def is_intron(self, p, strict = False): #If in intron
    if not self.is_contain(p, strict = strict): return False
    l=list(range(1,self.blockCount))
    if self.is_reverse() : l=l[::-1]
    p1=p-self.start
    for i in l:
      if self.blockStarts[i] >= p1 >= self.blockStop(i-1):
        if strict :
          if self.strand != '-' and p1 == self.blockStarts[i]: return False
          if self.strand == '-' and p1 == self.blockStop(i-1): return False
        return True
    else: return False

def bed3_iter(file, bin = False):
  for l in file:
    try: yield Bed3(l, bin)
    except ValueError: pass


def bed6_iter(file, bin = False):
  for l in file:
    try: yield Bed6(l, bin)
    except ValueError: pass

def bed12_iter(file, bin = False, filt = [], chrs = None, verbose = False):
  dfilt = {}
  if len(filt) > 0: 
    for k in filt : dfilt[k] = 1
  chr = ''
  for l in file:
    try: b = Bed12(l, bin)
    except ValueError: continue
    if b.chr != chr :
      if chrs is not None and b.chr not in chrs : continue
      if verbose : print(b.chr)
      chr = b.chr
    if len(filt) > 0 and b.id not in dfilt: continue
    yield b
    
def bed12SelectIter(file, select = None, **kwargs):
  for b in bed12_iter(file, **kwargs):
    if select is None : yield b
    else : 
      if select(b) is not None : yield b
        
def bed12_fetch(file, id, bin = False, verbose = False):
  for b in bed12_iter(file, bin = bin, verbose = verbose) :
    if b.id == id : return b


class refGene(Bed12): # start with bins
  def __init__(self, x):
    l = x.strip().split('\t')
    self.exonStarts = com2tup(l[9])
    self.exonEnds = com2tup(l[10])
    start = int(l[4])
    blockSizes = [self.exonEnds[i] - self.exonStarts[i] for i in range(len(self.exonStarts))]
    blockStarts = [self.exonStarts[i] - start for i in range(len(self.exonStarts))]
    l2 = [l[2],l[4],l[5],l[1],l[11],l[3],l[6],l[7],'0',l[8],tup2com(blockSizes),tup2com(blockStarts)]
    Bed12.__init__(self, l2)
    self.bin = int(l[0])
    self.gid = self.name2 = l[12]
    self.exonFrames = com2tup(l[15])
    self.cdsStartStat = l[13]
    self.cdsEndStat = l[14]
    
  def __repr__(self): 
    return 'refGene object:\n'+str(self)+'\n'
    
def refGene_iter(file):
  for l in file:
    try: yield refGene(l)
    except ValueError: pass
    
def gpd_iter(file, filt = [], chrs = None, verbose = False): #refFlat_iter(file):
  '''GenePred with Gene Names
  '''
  dfilt = {}
  if len(filt) > 0: 
    for k in filt : dfilt[k] = 1
  chr = ''
  for l in file:
    lst = l.strip().split()
    starts = com2tup(lst[9])
    stops = com2tup(lst[10])
    st1 = tuple(map(lambda x: x - int(lst[4]), starts))
    sizes = tuple(map(lambda x, y: y - x, starts, stops))
    lstb = [lst[2],lst[4],lst[5],lst[1],lst[0],lst[3],lst[6],lst[7],"0,0,0",lst[8],sizes,st1]
    bed = Bed12(lstb)
    bed.symbol = bed.gid = bed.score ##
    if bed.chr != chr : 
      if chrs is not None and bed.chr not in chrs : continue
      if verbose : print(bed.chr)
      chr = bed.chr
    if len(filt) > 0 and bed.id not in dfilt: continue
    yield bed
refFlat_iter = gpd_iter

class gpdGene:
  '''genepred gene
  '''
  def __init__(self, id = '', b = None):
    if b is None :
      self.trans = []
      self.chr, self.start, self.stop, self.strand = '', None, None, ''
      self.id = id
    else : 
      self.trans = [b]
      self.chr, self.start, self.stop, self.strand = b.chr, b.start, b.stop, b.strand
      self.id = b.gid
  def add_trans(self, tr):
    if self.strand in ('+', '-') and tr.strand != self.strand :
      print('Inconsistent trans strand: {} {} {} {}'.format(tr.gid, tr.tid, tr.strand, self.strand))
      self.strand = '.'
    elif self.strand == '': self.strand = tr.strand
    self.trans.append(tr)
  def __repr__(self):
    s = "gene_id " + self.id + ', ' + str(len(self.trans)) + " transcripts, " + Exon.__repr__(self)
    for t in self.trans:
      s += '\n\t' + t.__repr__()
    return s
  def check(self):
    if len(self.trans) > 0 :
      if self.chr == '' : self.chr = self.trans[0].chr
      if self.trand == '' : self.strand = self.trans[0].strand
      if self.start is None : self.start, self.stop = self.trans[0].start, self.trans[0].stop
    for t in self.trans:
      #t.check()
      if t.start < self.start: self.start = t.start
      if t.stop > self.stop: self.stop = t.stop
  def merge_trans(self):
    '''generate a new transcript that merge all transcript exons in the gene
    '''
    es = []
    merge = []
    for t in self.trans:
      es += t.exons
    es.sort()
    me = es[0]
    for i in range(1, len(es)):
      if me.stop >= es[i].start:
        me = union(me, es[i])[0]
      else: 
        merge.append(me)
        me = es[i]
    merge.append(me)
    #merge.check()
    return exons2bed12(merge)
def selectMaxCDS(trans): 
  '''select transcript with max CDS length
  '''
  maxlen, mt = 0, None
  for t in trans:
    cds1, cds2 = t.cds_start(cdna = True), t.cds_stop(cdna = True) 
    try : cdslen = cds2 - cds1 
    except : continue
    if cdslen % 3 != 0 : continue
    if cdslen > maxlen : maxlen, mt = cdslen, t
  return mt

def gpdGeneIter(file, **kwargs):
  chr, gt = '', {}
  for b in gpd_iter(file, **kwargs):
    if b.chr != chr :
      for gid in gt : 
        yield gt[gid]
        #t = selectMaxCDS(gt[gid])
        #if t is not None : yield t
      chr, gt = '', {}
    if b.gid not in gt : gt[b.gid] = gpdGene(b = b)
    else : gt[b.gid].add_trans(b)
  for gid in gt : yield gt[gid]
    #t = selectMaxCDS(gt[gid])
    #if t is not None : yield t


def gpdSelectIter(file, select = selectMaxCDS, **kwargs):
  chr, gt = '', {}
  for b in gpd_iter(file, **kwargs):
    if b.chr != chr :
      for gid in gt : 
        t = selectMaxCDS(gt[gid])
        if t is not None : yield t
      chr, gt = '', {}
    if b.gid not in gt : gt[b.gid] = []
    gt[b.gid].append(b)
  for gid in gt : 
    t = selectMaxCDS(gt[gid])
    if t is not None : yield t
      
def gpd_fetch(file, id):
  for b in gpd_iter(file):
    if b.id == id : return b

def short_bed(s, name = ''): #like chr1:1-200:+
  lst = s.strip().split(":")
  l1 = lst[1].split("-")
  if name == "": name = s.strip()
  if len(lst) >= 3: strand = lst[2]
  else: stand = '.'
  return Bed6([lst[0], l1[0], l1[1], name, 0, strand])

def exons2bed12(exons, id = 'exons_merge', thick_start = None, thick_stop = None) : #From exons to bed12
  exons.sort()
  chr, start, stop = exons[0].chr, exons[0].start, exons[-1].stop
  strand, score = exons[0].strand, exons[0].score
  if thick_start is None or thick_stop is None :
    thick_start = thick_stop = start
  itemRgb = 0
  blockCount = len(exons)
  blockSizes, blockStarts = [], []
  for e in exons:
    blockSizes.append(len(e))
    blockStarts.append(e.start - start)
  return Bed12([chr, start, stop, id, score, strand, thick_start, thick_stop, itemRgb, blockCount, tuple(blockSizes), tuple(blockStarts)])

def sub(a, b): # a - b
  if a.chr != b.chr : return [a]
  if a.start >= b.stop or a.stop <= b.start : return [a]
  out = []
  if a.start < b.start : out.append(a(stop = b.start))
  if b.stop < a.stop: out.append(a(start = b.stop))
  return out

def intersect(a, b): # a & b
  out = []
  if a.chr != b.chr : return out
  if a.start >= b.stop or a.stop <= b.start : return out
  out.append(a(start = max(a.start, b.start), stop = min(a.stop, b.stop)))
  return out

def union(a, b): # a U b
  if a.chr != b.chr or a.start > b.stop or a.stop < b.start :
    return [a, b]
  return [a(start = min(a.start, b.start), stop = max(a.stop, b.stop))]

def test():
  return 1





'''
Some useful tools
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''

Nt=['A','C','G','T']

def rc(seq):
  '''reverse complement
  '''
  comps = {'A':"T", 'C':"G", 'G':"C", 'T':"A",
       'B':"V", 'D':"H", 'H':"D", 'K':"M",
       'M':"K", 'R':"Y", 'V':"B", 'Y':"R",
       'W':'W', 'N':'N', 'S':'S'}
  return ''.join([comps[x] for x in seq.upper()[::-1]])

def overlap(A, B):
  '''
  if A is overlapping with B.
  '''
  if(A.chr != B.chr) : return False
  if (A.stop <= B.start) : return False
  if (B.stop <= A.start) : return False
  return True

def inside(A, B):
  '''
  if A is inside B.
  '''
  if(A.chr != B.chr) : return False
  if (B.stop < A.stop) : return False
  if (A.start < B.start) : return False
  return True
  
def distance(A,B):
  if A.chr!=B.chr: return None
  if overlap(A,B): return 0
  if A.start<B.start : return B.start-A.stop
  else: return A.start-B.stop
  
def centerin(A,B):
  '''
  If center of A is in B
  '''
  if overlap(A,B):
    return B.start<=A.center()<=B.stop
  else: return False

def end5(A):
  if A.strand=='-':
    return A.stop
  else:
    return A.start

def end3(A):
  if A.strand=='-':
    return A.start
  else:
    return A.stop


def fa_iter(file): # also in fa.py
  id = sq = ""
  for l in file:
    l = l.strip()
    if l == '' : continue
    if l[0] == '>':
      if id != '':
        yield (id,sq)
      sq = ''
      id = l[1:]
    else:
      sq += l.replace('U','T').replace('u','t')
  yield (id,sq)

def cover_iter(bedIter, weight= lambda x: 1): 
  '''bed coverage
  '''
  current = []
  chr = ""
  start = 0
  clen = 0
  for bed in bedIter:
    bed.weight = weight(bed)
    #print bed
    if clen == 0 :
      chr = bed.chr
      start = bed.start
      current.append(bed)
      clen = 1
    elif chr == bed.chr:
      if start == bed.start: pass
      else:
        assert start < bed.start, "Records must be sorted!\n"
        while clen > 0 and current[0].stop <= bed.start:
          cover = 0
          for i in range(clen):
            cover += current[i].weight
          stop = current[0].stop
          yield (chr, start, stop, cover)
          #print clen
          start = stop
          i = 0
          while i < clen and current[i].stop <= stop:
            i += 1
          current[0:i] = []
          clen = len(current)
          #print "clean", i, clen
        if clen > 0 and current[0].stop > bed.start:
          cover = 0
          for i in range(clen):
            cover += current[i].weight
          stop = bed.start
          yield (chr, start, stop, cover)
          #print clen
          start = stop
      i = clen - 1
      while i >= 0 and bed.stop < current[i].stop:
        i -= 1
      current[i+1:i+1] = [bed]
      clen = len(current)
      if clen == 1: start = bed.start
    else:
      while clen > 0 :
        cover = 0
        for i in range(clen):
          cover += current[i].weight
        stop = current[0].stop
        yield (chr, start, stop, cover)
        start = stop
        i = 0
        while i < clen and current[i].stop <= stop:
          i += 1
        #for i in range(clen):
          #if current[i].stop > stop: break
        current[0:i] = []
        clen = len(current)
      chr = bed.chr
      start = bed.start
      current.append(bed)
      clen = 1
  
  while clen > 0 :
    cover = 0
    for i in range(clen):
      cover += current[i].weight
    stop = current[0].stop
    yield (chr, start, stop, cover)
    start = stop
    i = 0
    while i < clen and current[i].stop <= stop:
      i += 1
    #for i in range(clen):
      #if corrent[i].stop > stop: break
    current[0:i] = []
    clen = len(current)
    
def overlap_iter(bedIterA, bedIterB, func=overlap, ignoreStrand = True, counts = [0,0]): ### Need Revise!!
  '''overlap of two bed iteraters (files), both should be sorted.
  '''
  lst = []
  ac = next(bedIterA) # .next() current bed A
  counts[0] += 1
  Aend = False
  for b in bedIterB:
    #print len(lst)
    counts[1] += 1
    j = -1
    cut = True
    for i in range(len(lst)):
      if lst[i].chr < b.chr: j = i
      elif lst[i].chr == b.chr:
        if cut and (lst[i].stop > b.start or func(lst[i], b)):
          j = i - 1
          cut = False          
        if func(lst[i], b):
          if ignoreStrand or lst[i].strand == b.strand : yield (lst[i], b)
    lst[0:(j+1)] = []
    if ac.chr > b.chr or (ac.chr == b.chr and ac.start > b.stop) : continue
    if func(ac, b):
      if ignoreStrand or ac.strand == b.strand : yield (ac, b)
    if Aend == False : lst.append(ac)
    c = 0
    for a in bedIterA:
      counts[0] += 1
      assert a.chr > ac.chr or a.start >= ac.start, "Records must be sorted!\n"+str(a)
      ac = a
      c += 1
      if a.chr < b.chr: continue
      elif a.chr == b.chr:
        if a.stop < b.start: continue
        if func(a, b):
          if ignoreStrand or a.strand == b.strand : yield (a, b)
        elif a.start > b.stop: break
      else: break
      lst.append(a)
    if c == 0 : Aend = True
    #print len(lst)

def rand_overlap_iter(bedIterA, bedListB, func=overlap, ignoreStrand = True): # ListB should have no overlap
  '''overlap of two bed iteraters (files), A is random, B is list, should be sorted and have no overlap.
  '''
  m = len(bedListB)
  for a in bedIterA:
    i0 = 0
    i1 = m-1
    #print a
    while i1 - i0 > 1 :
      i = (i1 + i0) // 2
      #print i, i0, i1
      if a > bedListB[i] : i0 = i
      else: i1 = i
    i = i0
    while func(a, bedListB[i]):
      if ignoreStrand or a.strand == bedListB[i].strand :
        yield (a, bedListB[i])
      i -= 1
    i = i1
    while func(a, bedListB[i]):
      if ignoreStrand or a.strand == bedListB[i].strand :
        yield (a, bedListB[i])
      i += 1
      #else : 
        #print 'i0', ignoreStrand, a.id, a.strand, bedListB[i0].id, bedListB[i0].strand
    #if func(a, bedListB[i1]):
      #if ignoreStrand or a.strand == bedListB[i1].strand :
        #yield (a, bedListB[i1])
      #else : 
        #print 'i1', ignoreStrand, a,id, a.strand, bedListB[i1].id, bedListB[i1].strand

def find_overlap(q, lst, func=overlap, ignoreStrand = True): 
  '''List should have no overlap
  '''
  m = len(lst)
  i0 = 0
  i1 = m-1
  while i1 - i0 > 1 :
    i = (i1 + i0) / 2
    if q > lst[i] : i0 = i
    else: i1 = i
  if func(q, lst[i0]):
    if ignoreStrand or q.strand == lst[i0].strand :
      return i0
  if func(q, lst[i1]):
    if ignoreStrand or q.strand == lst[i0].strand :
      return i1
  return -1

def bed2seq(seq, bed):
  if type(seq) == str : idx = False
  else : idx = True
  s = ''
  for e in bed.exons:
    if idx : es = seq.fetch(e.chr, start = e.start, stop = e.stop)
    else : es = seq[e.start:e.stop]
    if e.strand == '-':
      es = rc(es)
    s += es
  return s

def gtf2seq(seq, gtftrans):
  if type(seq) == str : idx = False
  else : idx = True
  s = ''
  for e in gtftrans.exons:
    if idx : es = seq.fetch(e.chr, start = e.start, stop = e.stop)
    else : es = seq[e.start:e.stop]
    if e.strand == '-':
      es = rc(es)
    s += es
  return s

def trans2seq(seq, trans):
  if type(seq) == str : idx = False
  else : idx = True
  s = ''
  for e in trans.exons:
    if idx : es = seq.fetch(e.chr, start = e.start, stop = e.stop)
    else : es = seq[e.start:e.stop]
    if e.strand == '-':
      es = rc(es)
    s += es
  return s


def exon2seq(seq, e):
  if type(seq) == str : s = seq[e.start:e.stop]
  else : s = seq.fetch(e.chr, start = e.start, stop = e.stop)
  if e.strand == '-':
    s = rc(s)
  return s

def cdna_pos(trans, p):
  exons = trans.exons
  exons.sort(reverse = trans.is_reverse())
  if p < trans.start or p > trans.stop:
    return None
  #p1 = p - trans.start
  pos = 0
  for e in exons:
    if e.start <= p <= e.stop:
      if trans.is_reverse():
        pos += e.stop - p
      else:
        pos += p - e.start
      return pos
    else:
      pos += len(e)
  return None

def genome_pos(trans, p, bias = 1):
  if p < 0 or p > trans.cdna_length():
    return None
  p1 = p
  pos = trans.start
  #l=range(len())
  for e in trans.exons:
    if len(e) - p1 >= bias:
      if trans.is_reverse():
        pos = e.stop - p1
      else:
        pos = p1 + e.start
      return pos
    else:
      p1 -= len(e)
  return None

def non_overlap_lists(lst):
  nolists=[]
  for b in lst:
    f = False
    for l in nolists:
      if overlap(b, l[-1]):
        l.append(b)
        f = True
        break
    if f : continue
    nolists.append([b])
  return nolists








'''
Open reading frame (ORF) processing
Copyright (c) 2016 Peng Zhang <zhpn1024@163.com>
'''

from . import fa

codonSize = 3
cstart = ['ATG']
cstartlike = ['TTG', 'CTG', 'GTG', 'AAG', 'AGG', 'ACG', 'ATT', 'ATA', 'ATC'] #AAG & AGG may be removed
#caltstart = cstartlike
cstop = ['TGA', 'TAA', 'TAG']

senseframes = [1, 2, 3]
antiframes = [-1, -2, -3]
frames = [1, 2, 3, -1, -2, -3]

class Orf:
  '''multiple start with same stop (in the same ORF)
  '''
  def __init__(self, lst = [], frame = 0, stop = -1):
    if len(lst) != 0:
      #lst = l.strip().split('\t')
      self.frame = int(lst[0])
      if '' == lst[1] : 
        self.starts = []
      else : self.starts = list(map(int, lst[1].split(',')))
      if lst[2] == '' : self.altstarts = []
      else : self.altstarts = list(map(int, lst[2].split(',')))
      self.stop = int(lst[3])
    else:
      self.frame = frame
      self.starts = []
      self.altstarts = []
      self.stop = stop
      self.has_stop_codon = True # For Non-stop end
      #self.end = self.stop - 3 
  def __str__(self):
    return "%d\t%s\t%s\t%d" % (self.frame, ','.join(map(str, self.starts)), ','.join(map(str, self.altstarts)), self.stop)
  def __repr__(self):
    return "Orf object: " + str(self)
  def has_start(self):
    return len(self.starts) + len(self.altstarts) > 0
  def has_stop(self):
    return self.stop >= 0
  def __len__(self):
    if not self.is_complete() : return 0
    if self.stop > 0 : return self.stop - self.start()
    #elif self.end > 0 : return self.end - self.start()
    else : return 0
  def __cmp__(self, other):
    return cmp(len(self), len(other)) or cmp(self.start(),other.start)
  def length(self):
    return len(self)
  def aa_len(self):
    l = len(self)
    if l == 0 : return 0
    if self.has_stop_codon : l -= 3
    return l // codonSize
  #@property
  def start(self, alt = True):
    if not self.has_start() : return None
    if alt : return min(self.starts + self.altstarts)
    else :
      if len(self.starts) > 0 : return min(self.starts)
      else : return None
  def has_strictstart(self):
    return len(self.starts) > 0
  def is_complete(self):
    return self.has_start() and self.has_stop()
  def orf_by_ribo(self, frame, start, stop, srange = 4): #look for TIS near ribo suggested ORF
    if self.frame != frame: return None
    if self.stop < start: return None
    if min(self.starts + self.altstarts) > stop : return None
    dmin = srange * codonSize
    sm = -1
    for s in self.starts :
      d = abs(s - start)
      if d < dm :
        dm = d
        sm = s
    if sm > 0 : return (s, self.stop)
    for s in self.altstarts :
      d = abs(s - start)
      if d < dm :
        dm = d
        sm = s
    if sm > 0 : return (s, self.stop)
    return None
  def filtByLen(self, minaalen, tail = -1):
    end = self.stop - 3
    if self.stop < 0 : 
      end = tail
      start = self.start()
      if start is None : return
      end -= (end - start) % 3 # keep in frame
      self.stop = end
      self.has_stop_codon = False
    #if stop < 0 : return
    th = minaalen * 3
    rm = False
    for i, s in enumerate(self.starts):
      if end - s < th : 
        rm = True
        break
    if rm : self.starts[i:] = []
    rm = False
    for i, s in enumerate(self.altstarts):
      if end - s < th : 
        rm = True
        break
    if rm : self.altstarts[i:] = []

def allorf(seq, strand = '+', minaalen = 0, tail = -1) :
  '''find all possible ORFs in the sequence
  '''
  seq = seq.upper().replace('U','T')
  if strand == '+' : fr = senseframes
  elif strand == '-' : 
    fr = antiframes
    antiseq = fa.rc(seq)
  else: 
    fr = frames
    antiseq = fa.rc(seq)
  
  length = len(seq)
  for f in fr:
    if f > 0: s = seq
      #fa = f
    else: s = antiseq
    fa = abs(f)
    o = Orf(frame = f)
    for i in range(fa-1, length, codonSize):
      try: codon = s[i:i+codonSize]
      except: break
      if codon in cstart: o.starts.append(i)
      elif codon in cstartlike: o.altstarts.append(i)
      if codon in cstop:
        #o.end = i
        o.stop = i + codonSize
        o.filtByLen(minaalen = minaalen, tail = tail)
        if o.has_start():
          yield o
        o = Orf(frame = f)
    o.filtByLen(minaalen = minaalen, tail = tail)
    if o.has_start() : yield o
def orflist(seq, strand = '+', sort = True, minaalen = 0, tail = -1):
  ol = list(allorf(seq, strand = strand, minaalen = minaalen, tail = tail))
  #for o in allorf(seq, strand = strand, minaalen = minaalen, tail = tail):
    #ol.append(o)
  if sort : ol.sort(reverse = True)
  return ol

class FixedOrf():
  '''only one start and one stop
  '''
  def __init__(self, start, stop = -1, has_stop_codon = True):
    self.start = start
    self.stop = stop
    self.has_stop_codon = has_stop_codon
  def __len__(self):
    if self.stop > 0 : return self.stop - self.start
    #elif self.end > 0 : return self.end - self.start
    else : return 0
  def length(self):
    return len(self)
  def __repr__(self):
    return "Fixed ORF object: %d - %d" % (self.start, self.stop)
  def frame(self): # 0 based
    return self.start % 3
  def __cmp__(self, other):
    return cmp(self.start, other.start) or cmp(self.stop, other.stop)
  def orfstr(self, seq):
    return "%s|%d-%d" % (seq[self.start:self.start+3], self.start, self.stop)
def orf_by_pos(seq, pos): ### Unkown start codon, only to find stop codon
  for i in range(pos, len(seq), codonSize):
    try: codon = seq[i:i+codonSize]
    except: FixedOrf(start = pos, stop = i, has_stop_codon = False)
    if codon in cstop: return FixedOrf(start = pos, stop = i + 3)
      #i += codonSize
      #break
  #o = FixedOrf(start = pos, stop = i)
  #return o
          
def orfDict(orflist, alt = True):
  '''dict of start -> stop
  '''
  od = {}
  for o in orflist:
    if o.stop < 0 : continue
    for s in o.starts: od[s] = o.stop
    if alt:
      for s in o.altstarts: od[s] = o.stop
  return od
def nearest_start(s, od, flank = 1):
  '''nearest start in ORF dict from s
  '''
  if s in od: return s
  for i in range(1, flank + 1):
    if s + i in od : return s + i ###
    if s - i in od : return s - i
  return None

codonTable = {'TTT':'F','TTC':'F','TTA':'L','TTG':'L', 'TCT':'S','TCC':'S','TCA':'S','TCG':'S', 'TAT':'Y','TAC':'Y','TAA':'*','TAG':'*', 'TGT':'C','TGC':'C','TGA':'*','TGG':'W', 
               'CTT':'L','CTC':'L','CTA':'L','CTG':'L', 'CCT':'P','CCC':'P','CCA':'P','CCG':'P', 'CAT':'H','CAC':'H','CAA':'Q','CAG':'Q', 'CGT':'R','CGC':'R','CGA':'R','CGG':'R', 
               'ATT':'I','ATC':'I','ATA':'I','ATG':'M', 'ACT':'T','ACC':'T','ACA':'T','ACG':'T', 'AAT':'N','AAC':'N','AAA':'K','AAG':'K', 'AGT':'S','AGC':'S','AGA':'R','AGG':'R', 
               'GTT':'V','GTC':'V','GTA':'V','GTG':'V', 'GCT':'A','GCC':'A','GCA':'A','GCG':'A', 'GAT':'D','GAC':'D','GAA':'E','GAG':'E', 'GGT':'G','GGC':'G','GGA':'G','GGG':'G', 
              }
AACodon = {'F':['TTC','TTT'], 'L':['CTG','CTC','CTT','TTG','TTA','CTA'], 'M':['ATG'], 'I':['ATC','ATT','ATA'], 'V':['GTG','GTC','GTT','GTA'],
           'S':['AGC','TCC','TCT','TCA','AGT','TCG'], 'P':['CCC','CCT','CCA','CCG'], 'T':['ACC','ACA','ACT','ACG'], 'A':['GCC','GCT','GCA','GCG'],
           'Y':['TAC','TAT'], '*':['TAA','TAG','TGA'], 'H':['CAC','CAT'], 'Q':['CAG','CAA'], 'N':['AAC','AAT'], 'K':['AAG','AAA'], 'D':['GAC','GAT'], 'E':['GAG','GAA'],
           'C':['TGC','TGT'], 'W':['TGG'], 'R':['AGA','AGG','CGG','CGC','CGA','CGT'], 'G':['GGC','GGA','GGG','GGT']
           } # ordered by human codon usage

allcodons = list(codonTable.keys())

def translate(seq):
  aa = ""
  for i in range(0, len(seq), codonSize):
    try : a = codonTable[seq[i:i+codonSize].upper().replace('U','T')]
    except : a = 'X'
    aa += a
  return aa

def senseMut(seq, mutkey, cds1 = 0, cds2 = None) : # pass
  if cds2 is None : cds2 = len(seq)
  if type(mutkey) == str : keys = [mutkey]
  else : keys = mutkey
  seq = seq.upper()
  change = {}
  has_key = True
  while has_key :
    has_key = False
    for key in keys : 
      p = seq.find(key)
      while p >= 0 : 
        has_key = True
        if p not in change : change[p] = 0
        mut = sense_mut1(seq, p, p+len(key), cds1, cds2, change[p])
        change[p] += 1
        if mut is None : 
          print('Cannot find sense mutation for {} at {} of {}!'.format(key, p, seq))
          return None
        newseq, i, c = mut
        print('{} {} changed to {}, {}'.format(i, seq[i:i+len(c)], c, newseq))
        seq = newseq #, i, c = mut
        p = seq.find(key)
  return seq

bases = ['T', 'C', 'A', 'G']
def sense_mut1(seq, start, stop, cds1, cds2, n = 0) : # mutate any one base in a region. n is time of retries.
  #incds = True
  if start >= cds2 or stop <= cds1 : # incds = False
    return base_mut(seq, start, stop, n)
  else : 
    n1 = (cds1 - start) * 4
    if n1 < 0 : n1 = 0
    if n < n1 : return base_mut(seq, start, cds1, n) # upstream mutation
    n2 = (stop - cds2) * 4
    if n2 < 0 : n2 = 0
    if n < n1 + n2 : return base_mut(seq, cds2, stop, n-n1) # downstream mutation
    return codon_mut(seq, start, stop, cds1, cds2, n-n1-n2) # codon mutation
########
def base_mut(seq, start, stop, n = 0) : 
  if n >= (stop - start) * 4 : return None # exceed limit
  i = int((start + stop) / 2)
  shift = n / 4 # 4 bases
  if shift % 2 > 0 : i -= int(shift / 2) + 1
  else : i += int(shift / 2)
  bi = n % 4
  newseq = seq[0:i] + bases[bi] + seq[i+1:]
  return newseq, i, bases[bi]

def codon_mut(seq, start, stop, cds1, cds2, n = 0) : # pass
  ni, ns = n, 0
  for i in range(cds1, cds2, 3) : 
    if i + 3 <= start : continue
    if i >= stop : return None
    codon = seq[i:i+3]
    aa = codonTable[codon]
    aac = AACodon[aa]
    if ni >= len(aac) : 
      ni -= len(aac)
      continue
    c2 = aac[ni]
    #for j in range(max(start-i,0), min(i+3-stop,3)) : 
    newseq = seq[0:i] + c2 + seq[i+3:]
    return newseq, i, c2

def is_start(seq, pos, alt = False, flank = 0):
  '''if start / alt start codon is nearby
  '''
  for i in range(pos - flank, pos + flank + 1):
    try : c = seq[i:i+codonSize]
    except : continue
    if c in cstart : return True
    elif alt and c in cstartlike : return True
  return False
def is_startlike(seq, pos, flank = 0):
  return is_start(seq, pos, alt = True, flank = flank)


